#coding: utf-8
from __future__ import division

import numpy as np
import scipy
import scipy.constants
from astropy.coordinates.angles import Angle
from astropy.time import Time
from numpy.testing import assert_allclose, assert_equal, assert_raises
from pyoperators import (
    Operator, Cartesian2SphericalOperator, CompositionOperator,
    BlockDiagonalOperator, IdentityOperator, MultiplicationOperator,
    Spherical2CartesianOperator, flags)
from pyoperators.utils import all_eq, isscalarlike, product
from pyoperators.utils.testing import (
    assert_is_instance, assert_is_type, assert_same)
from pysimulators.operators import (
    BlackBodyOperator, ConvolutionTruncatedExponentialOperator,
    PowerLawOperator, RollOperator, block_diagonal,
    CartesianEquatorial2GalacticOperator,
    CartesianEquatorial2HorizontalOperator,
    CartesianGalactic2EquatorialOperator,
    CartesianHorizontal2EquatorialOperator,
    SphericalEquatorial2GalacticOperator,
    SphericalEquatorial2HorizontalOperator,
    SphericalGalactic2EquatorialOperator,
    SphericalHorizontal2EquatorialOperator)


def test_partitioning_chunk():

    @block_diagonal('value', 'mykey', axisin=0)
    @flags.square
    class MyOp(Operator):
        def __init__(self, arg1, value, arg3, mykey=None, **keywords):
            Operator.__init__(self, **keywords)
            self.arg1 = arg1
            self.value = value
            self.arg3 = arg3
            self.mykey = mykey

        def direct(self, input, output):
            output[...] = self.value * input
        __str__ = Operator.__repr__

    @block_diagonal('value', 'mykey', axisin=0)
    @flags.square
    class MySupOp(Operator):
        def __init__(self, arg1, value, arg3, mykey=None, **keywords):
            Operator.__init__(self, **keywords)
            self.arg1 = arg1
            self.value = value
            self.arg3 = arg3
            self.mykey = mykey

    class MySubOp(MySupOp):
        def direct(self, input, output):
            output[...] = self.value * input
        __str__ = Operator.__repr__

    arg1 = [1, 2, 3, 4, 5]
    arg3 = ['a', 'b', 'c', 'd']

    def func(cls, n, v, k):
        n1 = 1 if isscalarlike(v) else len(v)
        n2 = 1 if isscalarlike(k) else len(k)
        nn = max(n1, n2) if n is None else 1 if isscalarlike(n) else len(n)
        if not isscalarlike(v) and not isscalarlike(k) and n1 != n2:
            # the partitioned arguments do not have the same length
            assert_raises(ValueError, lambda: cls(arg1, v, arg3, mykey=k,
                                                  partitionin=n))
            return
        if nn != max(n1, n2) and (not isscalarlike(v) or not isscalarlike(k)):
            # the partition is incompatible with the partitioned arguments
            return #XXX test assert_raises(ValueError)

        op = cls(arg1, v, arg3, mykey=k, partitionin=n)
        if nn == 1:
            v = v if isscalarlike(v) else v[0]
            k = k if isscalarlike(k) else k[0]
            func2(cls, op, v, k)
        else:
            assert op.__class__ is BlockDiagonalOperator
            assert len(op.operands) == nn
            if n is None:
                assert op.partitionin == nn * (None,)
                assert op.partitionout == nn * (None,)
                return
            v = nn * [v] if isscalarlike(v) else v
            k = nn * [k] if isscalarlike(k) else k
            for op_, v_, k_ in zip(op.operands, v, k):
                func2(cls, op_, v_, k_)
            expected = np.hstack(n_*[v_] for n_, v_ in zip(n, v))
            input = np.ones(np.sum(n))
            output = op(input)
            assert_equal(output, expected)

    def func2(cls, op, v, k):
        assert op.__class__ is cls
        assert op.arg1 is arg1
        assert op.value is v
        assert hasattr(op, 'arg3')
        assert op.arg3 is arg3
        assert op.mykey is k
        input = np.ones(1)
        output = op(input)
        assert_equal(output, v)

    for c in (MyOp, MySubOp):
        for n in (None, 2, (2,), (4, 2)):
            for v in (2., (2.,), (2., 3)):
                for k in (0., (0.,), (0., 1.)):
                    yield func, c, n, v, k


def test_partitioning_stack():

    @block_diagonal('value', 'mykey', new_axisin=0)
    @flags.square
    class MyOp(Operator):
        def __init__(self, arg1, value, arg3, mykey=None, **keywords):
            Operator.__init__(self, **keywords)
            self.arg1 = arg1
            self.value = value
            self.arg3 = arg3
            self.mykey = mykey

        def direct(self, input, output):
            output[...] = self.value * input
        __str__ = Operator.__repr__

    @block_diagonal('value', 'mykey', new_axisin=0)
    @flags.square
    class MySupOp(Operator):
        def __init__(self, arg1, value, arg3, mykey=None, **keywords):
            Operator.__init__(self, **keywords)
            self.arg1 = arg1
            self.value = value
            self.arg3 = arg3
            self.mykey = mykey

    class MySubOp(MySupOp):
        def direct(self, input, output):
            output[...] = self.value * input
        __str__ = Operator.__repr__

    arg1 = [1, 2, 3, 4, 5]
    arg3 = ['a', 'b', 'c', 'd']

    def func(cls, v, k):
        n1 = 1 if isscalarlike(v) else len(v)
        n2 = 1 if isscalarlike(k) else len(k)
        nn = max(n1, n2)
        if not isscalarlike(v) and not isscalarlike(k) and n1 != n2:
            # the partitioned arguments do not have the same length
            assert_raises(ValueError, lambda: cls(arg1, v, arg3, mykey=k))
            return

        op = cls(arg1, v, arg3, mykey=k)
        if nn == 1:
            v = v if isscalarlike(v) else v[0]
            k = k if isscalarlike(k) else k[0]
            func2(cls, op, v, k)
        else:
            assert op.__class__ is BlockDiagonalOperator
            assert len(op.operands) == nn
            v = nn * [v] if isscalarlike(v) else v
            k = nn * [k] if isscalarlike(k) else k
            for op_, v_, k_ in zip(op.operands, v, k):
                func2(cls, op_, v_, k_)
            input = np.ones(nn)
            output = op(input)
            assert_equal(output, v)

    def func2(cls, op, v, k):
        assert op.__class__ is cls
        assert op.arg1 is arg1
        assert op.value is v
        assert hasattr(op, 'arg3')
        assert op.arg3 is arg3
        assert op.mykey is k
        input = np.ones(1)
        output = op(input)
        assert_equal(output, v)

    for c in (MyOp, MySubOp):
        for v in (2., (2.,), (2., 3)):
            for k in (0., (0.,), (0., 1.)):
                yield func, c, v, k


def test_blackbody():

    def bb(w, T):
        c = 2.99792458e8
        h = 6.626068e-34
        k = 1.380658e-23
        nu = c/w
        return 2*h*nu**3/c**2 / (np.exp(h*nu/(k*T))-1)

    w = np.arange(90., 111) * 1e-6
    T = 15.
    flux = bb(w, T) / bb(100e-6, T)
    ops = [BlackBodyOperator(T, wavelength=wave, wavelength0=100e-6)
           for wave in w]
    flux2 = [op(1.) for op in ops]
    assert all_eq(flux, flux2)

    w, T = np.ogrid[90:111, 15:20]
    w = w * 1.e-6
    flux = bb(w, T) / bb(100e-6, T)
    ops = [BlackBodyOperator(T.squeeze(), wavelength=wave[0],
                             wavelength0=100e-6)
           for wave in w]
    flux2 = np.array([op(np.ones(T.size)) for op in ops])
    assert all_eq(flux, flux2)


def test_convolution_truncated_exponential():
    tau = (2, 3, 0)
    r = ConvolutionTruncatedExponentialOperator(tau, shapein=(3, 10))
    a = np.ones((3, 10))
    b = r(a)
    assert np.allclose(a, b)

    a[:, 1:] = 0
    b = r(a)
    assert_same(b[:2, :9], [np.exp(-np.arange(9) / _) for _ in tau[:2]])
    assert_same(b[2], a[2])
    assert_same(r.T.todense(), r.todense().T)
    assert_same(r.T.todense(inplace=True), r.todense(inplace=True).T)


def test_convolution_truncated_exponential_morphing():
    def func1(tau):
        assert_is_type(ConvolutionTruncatedExponentialOperator(tau),
                       IdentityOperator)
    for tau in 0, [0], (0,), np.array(0), np.array([0, 0]):
        yield func1, tau

    def func2(tau):
        assert_is_type(ConvolutionTruncatedExponentialOperator(tau),
                       ConvolutionTruncatedExponentialOperator)
    for tau in [0, 1], (0, 1), np.array([0, 1]):
        yield func2, tau


def test_power_law():
    c = scipy.constants.c
    nu = c / 120e-6
    nu0 = c / 100e-6
    values = np.arange(11)

    def func(alpha):
        expected = (nu/nu0)**alpha * values
        op = PowerLawOperator(alpha, nu, nu0)
        assert all_eq(op(values), expected)
    for alpha in (-1, np.arange(11)-5):
        yield func, alpha

    def func2(cls):
        op = cls([2, PowerLawOperator(1, nu, nu0)])
        assert_is_instance(op, PowerLawOperator)
    for cls in CompositionOperator, MultiplicationOperator:
        yield func2, cls


def test_roll():
    shape = np.arange(2, 6)
    v = np.arange(2*3*4*5).reshape(shape)
    for n in range(4):
        for axis in ((0,), (1,), (2,), (3,), (0, 1), (0, 2), (0, 3), (1, 2),
                     (1, 3), (2, 3), (0, 1, 2), (0, 1, 3), (0, 2, 3),
                     (1, 2, 3), (1, 2, 3), (0, 1, 2, 3)):
            expected = v.copy()
            for a in axis:
                expected = np.roll(expected, n, a)
            result = RollOperator(axis=axis, n=n)(v)
            yield assert_equal, result, expected


def test_equ2gal():
    equ2gal = CartesianEquatorial2GalacticOperator()
    gal2equ = CartesianGalactic2EquatorialOperator()

    assert equ2gal.I is equ2gal.T
    assert gal2equ.I is gal2equ.T
    assert_same(equ2gal.todense(shapein=3), equ2gal.data)
    assert_same(gal2equ.todense(shapein=3), equ2gal.data.T)

    shapes = (3, (2, 3), (4, 2, 3))

    def func(op, shape):
        vec = np.arange(product(shape)).reshape(shape)
        vec_ = vec.reshape((-1, 3))
        expected = np.empty(shape)
        expected_ = expected.reshape((-1, 3))
        for i in range(expected.size // 3):
            expected_[i] = op(vec_[i])
        actual = op(vec)
        assert_same(actual, expected)

    for op in (equ2gal, gal2equ):
        for shape in shapes:
            yield func, op, shape

    assert_is_instance(equ2gal(gal2equ), IdentityOperator)
    assert_is_instance(gal2equ(equ2gal), IdentityOperator)


def test_equ2hor():
    lat = 52
    lon = -64
    date = Time('1980-04-22 14:36:51.67', scale='ut1')
    E2h = CartesianEquatorial2HorizontalOperator
    gst = E2h._jd2gst(date.jd)
    lst = E2h._gst2lst(gst, lon)

    # Duffett-Smith §21
    assert_allclose(gst, 4.668119)
    assert_allclose(lst, 0.401453, rtol=1e-6)

    ra = Angle(lst - 5.862222, unit='hour').radian
    dec = Angle((23, 13, 10), unit='degree').radian
    s2c = Spherical2CartesianOperator('azimuth,elevation')
    op = CartesianEquatorial2HorizontalOperator('NE', date, lat, lon)
    incoords = s2c([ra, dec])
    outcoords = op(incoords)
    assert_same(op.I(outcoords), incoords)
    az, el = np.degrees(s2c.I(outcoords))

    # Duffett-Smith §25
    assert_allclose(az % 360, 283.271027)
    assert_allclose(el, 19.334345, rtol=1e-6)


def test_spherical():
    ops = ((SphericalEquatorial2GalacticOperator,
            CartesianEquatorial2GalacticOperator),
           (SphericalEquatorial2HorizontalOperator,
            CartesianEquatorial2HorizontalOperator),
           (SphericalGalactic2EquatorialOperator,
            CartesianGalactic2EquatorialOperator),
           (SphericalHorizontal2EquatorialOperator,
            CartesianHorizontal2EquatorialOperator))
    dirs_za = ((0, 0), (20, 0), (130, 0), (10, 20), (20, 190),
               ((0, 0), (20, 0), (130, 0), (10, 20), (20, 130)),
               (((0, 0), (20, 200), (130, 300)),))
    dirs_az = ((0, 0), (0, 20), (0, 130), (20, 10), (190, 20),
               ((0, 0), (0, 20), (0, 130), (20, 10), (130, 20)),
               (((0, 0), (200, 20), (300, 130)),))
    dirs_ea = ((90, 0), (70, 0), (-40, 0), (80, 20), (70, 190),
               ((90, 0), (70, 0), (-40, 0), (80, 20), (70, 130)),
               (((90, 0), (70, 200), (-40, 300)),))
    dirs_ae = ((0, 90), (0, 70), (0, -40), (20, 80), (190, 70),
               ((0, 90), (0, 70), (0, -40), (20, 80), (130, 70)),
               (((0, 90), (200, 70), (300, -40)),))
    shapes = ((), (), (), (), (), (5,), (1, 3))

    def func(cls_sph, cls_car, cin, cout, v, s, d):
        if 'Horizontal' in str(cls_sph):
            args = ('NE', Time('1980-04-22 14:36:51.67', scale='ut1'),
                    100.1, -80)
        else:
            args = ()
        op_sph = cls_sph(*args, conventionin=cin, conventionout=cout,
                         degrees=d)
        actual = op_sph(v)
        assert_equal(actual.shape, s + (2,))
        if d:
            v = np.radians(v)
        expected = Cartesian2SphericalOperator(cout)(
            cls_car(*args)(Spherical2CartesianOperator(cin)(v)))
        if d:
            np.degrees(expected, expected)
        assert_same(actual, expected)

    for cls_sph, cls_car in ops:
        for cin, vs in (('zenith,azimuth', dirs_za),
                        ('azimuth,zenith', dirs_az),
                        ('elevation,azimuth', dirs_ea),
                        ('azimuth,elevation', dirs_ae)):
            for cout in ('zenith,azimuth',
                         'azimuth,zenith',
                         'elevation,azimuth',
                         'azimuth,elevation'):
                for v, s in zip(vs, shapes):
                    for d in (False, True):
                        yield func, cls_sph, cls_car, cin, cout, v, s, d



from __future__ import absolute_import, division, print_function

import astropy.io.fits as pyfits
import numpy as np
from numpy.testing import assert_allclose
from pyoperators import DiagonalOperator, PackOperator, pcg
from pyoperators.utils.testing import assert_same, assert_eq
from pysimulators import Map
from pysimulators.interfaces.madmap1 import MadMap1Observation
from pysimulators.interfaces.madmap1.acquisition import _read_filters


def mapper_naive(tod, H):
    m, c = H.T(tod), H.T(np.ones_like(tod))
    m /= c
    m.coverage = c
    return m

profile = None #'test_madcap.png'
path = 'test/data/madmap1/'
map_ref = pyfits.open(path + 'naivemapSpirePSW.fits')['image'].data
name = 'SPIRE/PSW'
obs = MadMap1Observation(name, 135, path + 'todSpirePSW_be',
                         path + 'invnttSpirePSW_be',
                         path + 'madmapSpirePSW.fits[coverage]',
                         bigendian=True, missing_value=np.nan)

tod = obs.get_tod(unit='Jy/beam')
projection = obs.get_operator()
packing = PackOperator(np.isfinite(map_ref))

model = projection * packing
map_naive_1d = mapper_naive(tod, projection)
map_naive_2d = mapper_naive(tod, model)


def test_naive1():
    assert_same(map_naive_1d, map_ref[np.isfinite(map_ref)])


def test_naive2():
    assert_same(map_naive_2d, map_ref)


M = DiagonalOperator(packing(1/map_naive_2d.coverage))
assert np.all(np.isfinite(M.data))

invntt = obs.get_invntt_operator(fftw_flag='FFTW_PATIENT')


class Callback:
    def __init__(self):
        self.niterations = 0

    def __call__(self, x):
        self.niterations += 1

callback = Callback()
#callback = None


def test_ls():
    H = projection
    m = pcg(H.T * invntt * H, (H.T*invntt)(tod), M=M, tol=1e-7)
    map_ls_packed = Map(m['x'])
    map_ls_packed.header['TIME'] = m['time']

    #print('Elapsed time:', map_ls_packed.header['TIME'])
    #from tamasis import mapper_ls
    #map_ls_packed = mapper_ls(tod, projection, invntt=invntt, tol=1e-7, M=M,
    #                          callback=callback, criterion=False,
    #                          profile=profile)
    if profile:
        return
    print('Elapsed time:', map_ls_packed.header['TIME'])
    assert m['nit'] < 50
    ref = packing(Map(path + 'madmapSpirePSW.fits'))
    assert_allclose(map_ls_packed, ref, atol=1e-5)


def test_endian():
    bendian = _read_filters(path + 'invntt_be', bigendian=True)
    lendian = _read_filters(path + 'invntt_le')
    assert len(bendian) == 6
    assert bendian[0]['data'].size == 101
    assert_same(bendian[0]['data'][0], 5597147.4155586753)
    assert_eq(lendian, bendian)


if __name__ == '__main__':
    test_ls()



from pyoperators.utils.mpi import MPI

# force gfortran's read statement to always use the dot sign as fraction
# separator (PR47007)
import locale
locale.setlocale(locale.LC_NUMERIC, 'POSIX')
del locale

from . import _flib
from . import geometry
from .operators import *
from .datatypes import *
from .packedtables import *
from .instruments import *
#from .mpiutils import *
from .acquisitions import *
#from .processing import *
from .quantities import *
from .datautils import *
from .sparse import *
from .wcsutils import *

__version__ = '1.1.dev53'



# Copyrights 2010-2013 Pierre Chanial
# All rights reserved
#
from __future__ import absolute_import, division, print_function
from astropy.io import fits as pyfits
from astropy.wcs import WCS
from pyoperators import Operator
from pyoperators.flags import inplace, real, square
from pyoperators.utils import isscalarlike, operation_assignment
from . import _flib as flib
import astropy.wcs
import numpy as np
import scipy.interpolate as interp

__all__ = [
    'angle_lonlat',
    'barycenter_lonlat',
    'create_fitsheader',
    'create_fitsheader_for',
    'fitsheader2shape',
    'str2fitsheader',
    'DistortionOperator',
    'RotationBoresightEquatorialOperator',
    'WCSToPixelOperator',
    'WCSToWorldOperator',
]

# Currently, astropy wcs transform only handle input of ndim equal to 2
# See https://github.com/astropy/astropy/issues/976
ASTROPY_WCS_NDIM_IS_2 = True


def angle_lonlat(lon1, lat1, lon2=None, lat2=None):
    """
    Returns the angle between vectors on the celestial sphere in degrees.

    Parameters
    ----------
    lon1, lon2 : array of numbers
        longitude in degrees
    lat1, lat2 : array of numbers
        latitude in degrees

    Example
    -------
    >>> ra1, dec1 = 30, 70
    >>> ra2, dec2 = 30.5, 70.5
    >>> angle_lonlat((ra1, dec1), (ra2, dec2))
    >>> angle_lonlat(ra1, dec1, ra2, dec2)

    """
    if lon2 is None and lat2 is None:
        lon2, lat2 = lat1
        lon1, lat1 = lon1
    lon1 = np.array(lon1, float, ndmin=1, copy=False).ravel()
    lat1 = np.array(lat1, float, ndmin=1, copy=False).ravel()
    lon2 = np.array(lon2, float, ndmin=1, copy=False).ravel()
    lat2 = np.array(lat2, float, ndmin=1, copy=False).ravel()
    angle = flib.wcsutils.angle_lonlat(lon1, lat1, lon2, lat2)
    if angle.size == 1:
        angle = float(angle)
    return angle


def barycenter_lonlat(lon, lat):
    """
    Returns the barycenter of vectors on the celestial sphere.

    Parameters
    ----------
    lon : array of numbers
        Longitude in degrees.
    lat : array of numbers
        Latitude in degrees.
    """
    lon = np.array(lon, float, ndmin=1, copy=False).ravel()
    lat = np.array(lat, float, ndmin=1, copy=False).ravel()
    return flib.wcsutils.barycenter_lonlat(lon, lat)


def combine_fitsheader(headers, cdelt=None, pa=None):
    """
    Returns a FITS header which encompasses all input headers.

    Parameters
    ----------
    headers : list of pyfits.Header
        The headers of the maps that must be contained in the map of the
        returned header.
    cdelt : float or 2 element array
        Physical increment at the reference pixel. Default is the lowest
        header cdelt.
    pa : float
        Position angle of the Y=AXIS2 axis (=-CROTA2). Default is the mean
        header pa.
    """

    if not isinstance(headers, (list, tuple)):
        headers = (headers,)

    if len(headers) == 1 and cdelt is None and pa is None:
        return headers[0]

    # compute the world coordinates of the reference pixel of the combined map
    crval = barycenter_lonlat([h['CRVAL1'] for h in headers],
                              [h['CRVAL2'] for h in headers])

    # compute the combined map cdelt & pa
    if cdelt is None or pa is None:
        cdeltpa = [get_cdelt_pa(h) for h in headers]

    if cdelt is None:
        cdelt = min([np.min(abs(cdelt_)) for cdelt_, pa_ in cdeltpa])

    if pa is None:
        pa = flib.wcsutils.mean_degrees(
            np.array([pa_ for cdelt_, pa_ in cdeltpa]))

    # first, create the combined header with a single pixel centred on the
    # reference coordinates, with correct orientation and cdelt
    header0 = create_fitsheader((1, 1), cdelt=cdelt, pa=pa, crpix=(1, 1),
                                crval=crval)

    # then, 'enlarge' it to make it fit the edges of each header
    proj0 = WCS(header0)
    xy0 = []
    for h in headers:
        p = WCS(h)
        nx = h['NAXIS1']
        ny = h['NAXIS2']
        x = h['CRPIX1']
        y = h['CRPIX2']
        edges = p.wcs_pix2world(
            [0.5, x, nx+0.5, 0.5, x, nx+0.5, 0.5, x, nx+0.5],
            [0.5, 0.5, 0.5, y, y, y, ny+0.5, ny+0.5, ny+0.5], 1)
        xy0.append(proj0.wcs_world2pix(edges[0], edges[1], 1))

    xmin0 = np.round(min([min(c[0]) for c in xy0]))
    xmax0 = np.round(max([max(c[0]) for c in xy0]))
    ymin0 = np.round(min([min(c[1]) for c in xy0]))
    ymax0 = np.round(max([max(c[1]) for c in xy0]))

    header0['NAXIS1'] = int(xmax0 - xmin0 + 1)
    header0['NAXIS2'] = int(ymax0 - ymin0 + 1)
    header0['CRPIX1'] = 2 - xmin0
    header0['CRPIX2'] = 2 - ymin0

    return header0


def create_fitsheader(naxes=None, dtype=None, fromdata=None, extname=None,
                      crval=(0, 0), crpix=None, cunit='deg',
                      ctype=('RA---TAN', 'DEC--TAN'), cd=None, cdelt=None,
                      pa=None, radesys='ICRS', equinox=None):
    """
    Helper to create a FITS header.

    Parameters
    ----------
    naxes : array, optional
        (NAXIS1,NAXIS2,...) tuple, which specifies the data dimensions.
        Note that the dimensions in FITS and ndarrays are reversed.
    dtype : data-type, optional
        Desired data type for the data stored in the FITS file
    extname : None or string
        if a string is specified ('' can be used), the returned header
        type will be an Image HDU (otherwise a Primary HDU).
    crval : 2 element array, optional
        Reference pixel values (FITS convention).
    crpix : 2 element array, optional
        Reference pixel (FITS convention).
    ctype : 2 element string array, optional
        Projection types.
    cunit : string or 2 element string array
        Units of the CD matrix (default is degrees/pixel).
    cd : 2 x 2 array
        FITS parameters
            CD1_1 CD1_2
            CD2_1 CD2_2
    cdelt : float or 2 element array
        Physical increment at the reference pixel.
    pa : float
        Position angle of the Y=AXIS2 axis (=-CROTA2).
    radesys : string
        Coordinate reference frame for the RA and DEC axis columns. Default
        is 'ICRS'.
    equinox : float, optional
        Reference equinox (for non-ICRS reference frame).

    Example
    -------
    >>> from pysimulators import Map
    >>> map = Map.ones((10, 100), unit='Jy/pixel')
    >>> ny, nx = map.shape
    >>> map.header = create_fitsheader((nx, ny), cd=[[-1,0], [0,1]])

    """
    if naxes is not None and fromdata is not None or naxes is None and \
       fromdata is None:
        raise ValueError(
            "Either keyword 'naxes' or 'fromdata' must be specified.")

    if fromdata is None:
        naxes = np.array(naxes, dtype=int, ndmin=1)
        naxes = tuple(naxes)
        if len(naxes) > 8:
            raise ValueError('First argument is naxes=(NAXIS1,NAXIS2,...)')
        if dtype is not None:
            dtype = np.dtype(dtype)
            typename = dtype.name
        else:
            typename = 'float64'
    else:
        print("The keyword 'from_data' in create_fitsheader is deprecated and "
              "will be removed in an upcoming release. Use the function create"
              "_fitsheader_for instead.")
        array = np.array(fromdata, copy=False)
        naxes = tuple(reversed(array.shape))
        if dtype is not None:
            array = array.astype(dtype)
        if array.dtype.itemsize == 1:
            typename = 'uint8'
        elif array.dtype.names is not None:
            typename = None
        else:
            typename = array.dtype.name

    numaxis = len(naxes)

    header = pyfits.Header()
    if extname is None:
        header['simple'] = True  # primary header
    else:
        header['xtension'] = ('IMAGE', 'Image extension')
    if typename is not None:
        header['bitpix'] = (pyfits.PrimaryHDU.ImgCode[typename],
                            'Array data type')
    header['naxis'] = (numaxis, 'Number of array dimensions')
    for dim in range(numaxis):
        header['naxis' + str(dim+1)] = naxes[dim]
    if extname is None:
        header['extend'] = True
    else:
        header['pcount'] = (0, 'Number of parameters')
        header['gcount'] = (1, 'Number of groups')
        header['extname'] = extname

    if cd is not None:
        cd = np.asarray(cd, dtype=float)
        if cd.shape != (2, 2):
            raise ValueError('The CD matrix is not a 2x2 matrix.')
    else:
        if cdelt is None:
            return header
        cdelt = np.array(cdelt, float)
        if isscalarlike(cdelt):
            if ctype is None or len(ctype) == 2 and ctype[0] == 'RA---TAN' and\
               ctype[1] == 'DEC--TAN':
                cdelt = (-cdelt, cdelt)
            else:
                cdelt = (cdelt, cdelt)
        if pa is None:
            pa = 0.
        theta = np.deg2rad(-pa)
    cd = np.diag(cdelt).dot(np.array([[np.cos(theta), np.sin(theta)],
                                      [-np.sin(theta), np.cos(theta)]]))

    crval = np.asarray(crval, float)
    if crval.size != 2:
        raise ValueError('CRVAL does not have two elements.')

    if crpix is None:
        crpix = (np.array(naxes) + 1) / 2
    else:
        crpix = np.asarray(crpix, float)
    if crpix.size != 2:
        raise ValueError('CRPIX does not have two elements.')

    if len(ctype) != 2:
        raise ValueError('CTYPE does not have two elements.')

    if isscalarlike(cunit):
        cunit = (cunit, cunit)
    if len(cunit) != 2:
        raise ValueError('CUNIT does not have two elements.')

    header['crval1'] = crval[0]
    header['crval2'] = crval[1]
    header['crpix1'] = crpix[0]
    header['crpix2'] = crpix[1]
    header['cd1_1'] = cd[0, 0]
    header['cd2_1'] = cd[1, 0]
    header['cd1_2'] = cd[0, 1]
    header['cd2_2'] = cd[1, 1]
    header['ctype1'] = ctype[0]
    header['ctype2'] = ctype[1]
    header['cunit1'] = cunit[0]
    header['cunit2'] = cunit[1]

    if ctype[0][:3] == 'RA-' and ctype[1][:3] == 'DEC':
        header['radesys'] = radesys.upper()

    if equinox is not None:
        header['equinox'] = float(equinox)

    return header


def create_fitsheader_for(data, extname=None, crval=(0, 0), crpix=None,
                          ctype=('RA---TAN', 'DEC--TAN'), cunit='deg', cd=None,
                          cdelt=None, pa=None, radesys='ICRS', equinox=None):
    """
    Helper to create a FITS header from an ndarray.

    Parameters
    ----------
    data : array_like or None
        The array from which the dimensions and type will be extracted. Note
        that following the FITS convention, the dimension along X is the
        second value of the array shape and that the dimension along the
        Y axis is the first one.
    extname : None or string
        if a string is specified ('' can be used), the returned header
        type will be an Image HDU (otherwise a Primary HDU).
    crval : 2 element array, optional
        Reference pixel values (FITS convention).
    crpix : 2 element array, optional
        Reference pixel (FITS convention).
    ctype : 2 element string array, optional
        Projection types.
    cunit : string or 2 element string array
        Units of the CD matrix (default is degrees/pixel).
    cd : 2 x 2 array
        FITS parameters
            CD1_1 CD1_2
            CD2_1 CD2_2
    cdelt : float or 2 element array
        Physical increment at the reference pixel.
    pa : float
        Position angle of the Y=AXIS2 axis (=-CROTA2).
    radesys : string
        Coordinate reference frame for the RA and DEC axis columns. Default
        is 'ICRS'.
    equinox : float, optional
        Reference equinox (for non-ICRS reference frame).

    Example
    -------
    >>> from pysimulators import Map
    >>> map = Map.ones((10,100), unit='Jy/pixel')
    >>> map.header = create_fitsheader_for(map, cd=[[-1,0], [0,1]])

    >>> primary_header = create_fitsheader_for(None)
    >>> secondary_header = create_fitsheader_for(None, extname='secondary')

    """
    if data is None:
        naxes = ()
        dtype = np.int32
    else:
        data = np.array(data, ndmin=1)
        naxes = data.shape[::-1]
        if data.dtype.itemsize == 1:
            data = data.view('uint8')
        dtype = data.dtype
    return create_fitsheader(naxes, dtype=dtype,
                             extname=extname, crval=crval, crpix=crpix,
                             ctype=ctype, cunit=cunit, cd=cd, cdelt=cdelt,
                             pa=pa, radesys=radesys, equinox=equinox)


def fitsheader2shape(header):
    ndim = header['NAXIS']
    return tuple(header['NAXIS{0}'.format(i)] for i in range(ndim, 0, -1))


def get_cdelt_pa(header):
    """
    Extract scale and rotation from a FITS header.

    Parameter
    ---------
    header : pyfits Header
        The FITS header

    Returns
    -------
    cdelt : 2 element array
        Physical increment at the reference pixel.
    pa : float
        Position angle of the Y=AXIS2 axis (=-CROTA2).

    """
    try:
        cd = np.array([[header['CD1_1'], header['CD1_2']],
                       [header['CD2_1'], header['CD2_2']]], float)
    except KeyError:
        if any(not k in header for k in ('CDELT1', 'CDELT2', 'CROTA2')):
            raise KeyError('Header has no astrometry.')
        return (np.array([header['CDELT1'], header['CDELT2']]),
                -header['CROTA2'])

    det = np.linalg.det(cd)
    sgn = det / np.abs(det)
    cdelt = np.array([sgn * np.sqrt(cd[0, 0]**2 + cd[0, 1]**2),
                      np.sqrt(cd[1, 1]**2 + cd[1, 0]**2)])
    pa = -np.arctan2(-cd[1, 0], cd[1, 1])
    return cdelt, np.rad2deg(pa)


def has_wcs(header):
    """
    Returns True is the input FITS header has a defined World Coordinate
    System.
    """

    required = 'CRPIX,CRVAL,CTYPE'.split(',')
    keywords = np.concatenate([(lambda i: [r+str(i+1) for r in required])(i)
                               for i in range(header['NAXIS'])])
    return all([k in header for k in keywords])


def mean_degrees(array):
    """
    Returns the mean value of an array of values in degrees, by taking into
    account the discrepancy at 0 degree
    """
    return flib.wcsutils.mean_degrees(np.asarray(array, dtype=float).ravel())


def minmax_degrees(array):
    """
    Returns the minimum and maximum value of an array of values in degrees,
    by taking into account the discrepancy at 0 degree.
    """
    return flib.wcsutils.minmax_degrees(np.asarray(array, dtype=float).ravel())


def str2fitsheader(string):
    """
    Convert a string into a pyfits.Header object

    All cards are extracted from the input string until the END keyword is
    reached.
    """
    header = pyfits.Header()
    cards = header.ascard()
    iline = 0
    while (iline*80 < len(string)):
        line = string[iline*80:(iline+1)*80]
        if line[0:3] == 'END': break
        cards.append(pyfits.Card().fromstring(line))
        iline += 1
    return header


@real
@square
class DistortionOperator(Operator):
    """
    Distortion operator.

    The interpolation is performed using the Clough-Tocher method.
    Interpolation of points outside the object plane coordinates convex hull
    will return NaN.

    Parameters
    ----------
    xin : array of shape (npoints, ndims)
       Coordinates in the object plane.
    xout : array of shape (npoints, ndims)
       Coordinates in the image plane.

    """
    def __init__(self, xin, xout, **keywords):
        xin = np.asarray(xin)
        xout = np.asarray(xout)
        if xin.shape[-1] != 2:
            raise ValueError('The shape of the object plane coordinates should'
                             ' be (npoints,ndims), where ndims is only impleme'
                             'nted for 2.')
        if xin.shape != xout.shape:
            raise ValueError('The object and image coordinates do not have the'
                             ' same shape.')

        keywords['dtype'] = float
        Operator.__init__(self, **keywords)
        self.interp0 = interp.CloughTocher2DInterpolator(xin, xout[..., 0])
        self.interp1 = interp.CloughTocher2DInterpolator(xin, xout[..., 1])
        self.set_rule('I', lambda s: DistortionOperator(xout, xin))

    def direct(self, input, output):
        output[..., 0] = self.interp0(input)
        output[..., 1] = self.interp1(input)


@real
class RotationBoresightEquatorialOperator(Operator):
    """
    Convert coordinates in the instrument object plane into celestial
    coordinates, assuming a pointing direction and a position angle.

            input[1] along declination if PA=0
       \    |
        \PA_|
         \/ |
          \ |
           \|_______ input[0] along -R.A. if PA=0

    The routine is not accurate at the poles.
    Input coordinates are in arc seconds, pointing and output are in degrees.

    """
    def __init__(self, pointing, **keywords):
        if any(k not in pointing.dtype.names for k in ('ra', 'dec', 'pa')):
            raise TypeError('The input is not an equatorial pointing.')
        self.pointing = pointing
        Operator.__init__(self, **keywords)

    def direct(self, x, out):
        s = (Ellipsis,) + (x.ndim - 1) * (None,)
        ra = self.pointing[s].ra
        dec = self.pointing[s].dec
        pa = np.deg2rad(self.pointing[s].pa)
        cospa = np.cos(pa) / 3600
        sinpa = np.sin(pa) / 3600
        out[..., 1] = dec + (-x[..., 0] * sinpa + x[..., 1] * cospa)
        out[..., 0] = ra + (x[..., 0] * cospa + x[..., 1] * sinpa) /    \
                           np.cos(np.deg2rad(out[..., 1]))

    def reshapein(self, shape):
        return self.pointing.shape + shape


@real
@square
@inplace
class _WCSKapteynOperator(Operator):
    def __init__(self, wcs, **keywords):
        """
        wcs : FITS header or Kapteyn wcs.Projection instance
            Representation of the world coordinate system

        """
        import kapteyn.wcs as kwcs
        if isinstance(wcs, pyfits.Header):
            wcs = kwcs.Projection(wcs)
        if 'dtype' not in keywords:
            keywords['dtype'] = float
        wcs.allow_invalid = True
        Operator.__init__(self, **keywords)
        self.wcs = wcs

    def validatein(self, shapein):
        if len(shapein) == 0:
            raise ValueError('Invalid scalar input.')
        if shapein[-1] != 2:
            raise ValueError("Invalid last dimension: '{0}'.".format(
                             shapein[-1]))


class WCSKapteynToPixelOperator(_WCSKapteynOperator):
    """
    Operator for WCS world-to-pixel transforms.

    Example
    -------
    >>> header = pyfits.open('myfile.fits')
    >>> w = WCSKapteynToPixelOperator(header)
    >>> ra0, dec0 = 190, 77
    >>> pix = w([ra0, dec0])

    """
    def __init__(self, wcs, **keywords):
        _WCSKapteynOperator.__init__(self, wcs, **keywords)
        self.set_rule('I', lambda s: WCSKapteynToWorldOperator(s.wcs))
    __init__.__doc__ = _WCSKapteynOperator.__init__.__doc__

    def direct(self, input, output, operation=operation_assignment):
        operation(output, self.wcs.topixel(input))


class WCSKapteynToWorldOperator(_WCSKapteynOperator):
    """
    Operator for WCS pixel-to-world transforms.

    Example
    -------
    >>> header = pyfits.open('myfile.fits')
    >>> w = WCSKapteynToWorldOperator(header)
    >>> x0, y0 = 34, 65.2
    >>> radec = w([x0, y0])

    """
    def __init__(self, wcs, **keywords):
        _WCSKapteynOperator.__init__(self, wcs, **keywords)
        self.set_rule('I', lambda s: WCSKapteynToPixelOperator(s.wcs))
    __init__.__doc__ = _WCSKapteynOperator.__init__.__doc__

    def direct(self, input, output, operation=operation_assignment):
        operation(output, self.wcs.toworld(input))


@real
@square
@inplace
class _WCSOperator(Operator):
    def __init__(self, wcs, origin=0, **keywords):
        """
        wcs : FITS header or astropy.wcs.WCS instance
            Representation of the world coordinate system
        origin : int, optional
            Indexing convention : 0 for 0-based indexing such as C or numpy),
            1 for FITS or Fortran 1-based indexing.

        """
        if not isinstance(wcs, astropy.wcs.WCS):
            wcs = astropy.wcs.WCS(wcs)
        if 'dtype' not in keywords:
            keywords['dtype'] = float
        Operator.__init__(self, **keywords)
        self.wcs = wcs
        self.origin = origin

    def validatein(self, shapein):
        if len(shapein) == 0:
            raise ValueError('Invalid scalar input.')
        if shapein[-1] != 2:
            raise ValueError("Invalid last dimension: '{0}'.".format(
                             shapein[-1]))


class WCSToPixelOperator(_WCSOperator):
    """
    Operator for WCS world-to-pixel transforms.

    Example
    -------
    >>> header = pyfits.open('myfile.fits')
    >>> w = WCSToPixelOperator(header)
    >>> ra0, dec0 = 190, 77
    >>> pix = w([ra0, dec0])

    """
    def __init__(self, wcs, origin=0, **keywords):
        _WCSOperator.__init__(self, wcs, origin, classin=np.ndarray,
                              classout=np.ndarray, **keywords)
        self.set_rule('I', lambda s: WCSToWorldOperator(s.wcs, s.origin))
    __init__.__doc__ = _WCSOperator.__init__.__doc__

    def direct(self, input, output, operation=operation_assignment):
        if ASTROPY_WCS_NDIM_IS_2:
            input = input.reshape((-1, 2))
            output = output.reshape((-1, 2))
        operation(output, self.wcs.wcs_world2pix(input, self.origin))


class WCSToWorldOperator(_WCSOperator):
    """
    Operator for WCS pixel-to-world transforms.

    Example
    -------
    >>> header = pyfits.open('myfile.fits')
    >>> w = WCSToWorldOperator(header)
    >>> x0, y0 = 34, 65.2
    >>> radec = w([x0, y0])

    """
    def __init__(self, wcs, origin=0, **keywords):
        _WCSOperator.__init__(self, wcs, origin, classin=np.ndarray,
                              classout=np.ndarray, **keywords)
        self.set_rule('I', lambda s: WCSToPixelOperator(s.wcs, s.origin))
    __init__.__doc__ = _WCSOperator.__init__.__doc__

    def direct(self, input, output, operation=operation_assignment):
        if ASTROPY_WCS_NDIM_IS_2:
            input = input.reshape((-1, 2))
            output = output.reshape((-1, 2))
        operation(output, self.wcs.wcs_pix2world(input, self.origin))



from __future__ import division

import numpy as np
from numpy.testing import assert_allclose
from pyoperators.utils.testing import assert_same
from pysimulators.noises import (_fold_psd, _unfold_psd, _interp,
                                 _gaussian_psd_1f, _gaussian_sample)


def test_fold():
    p = [1, 1, 1, 1]
    p_folded = _fold_psd(p)
    assert_same(p_folded, [1, 2, 1])


def test_unfold():
    p = [1, 2, 1]
    p_unfolded = _unfold_psd(p)
    assert_same(p_unfolded, [1, 1, 1, 1])


def test_interp():
    x = [1, 2, 3, 5, 7]
    y = [[1, 2, 3, 5, 7], [0, 1, 2, 4, 6]]
    zs = ([-1, -0.5, 0, 0.5, 1, 1.5, 2, 2.5, 3, 3.5, 4, 4.5, 7.5, 8],
          [-1, -0.5], [2.25, 2.5], [8, 10])

    def func(z):
        out1 = _interp(z, x, y[0])
        assert_same(z, out1)
        out2 = _interp(z, x, y[1])
        out = _interp(z, x, y)
        assert_same(out, np.vstack([out1, out2]))
    for z in zs:
        yield func, z


def test_gaussian():
    fsamp = 5
    nsamples = 2**10
    white = [0.3, 0.5]
    fknee = 0.
    alpha = 1

    def func(twosided):
        np.random.seed(0)
        p = _gaussian_psd_1f(nsamples, fsamp, white, fknee, alpha,
                             twosided=twosided)
        t = _gaussian_sample(nsamples, fsamp, p, twosided=twosided)
        return p, t

    p1, t1 = func(False)
    p2, t2 = func(True)

    bandwidth = fsamp / nsamples
    assert_same(np.sum(p1, axis=-1) * bandwidth,
                np.sum(p2, axis=-1) * bandwidth, rtol=1000)

    assert_allclose(np.sqrt(np.sum(p1, axis=-1) * bandwidth), white, rtol=1e-3)
    assert_same(t1, t2)



# Copyrights 2010-2013 Pierre Chanial
# All rights reserved

from __future__ import absolute_import, division, print_function
from copy import copy
from pyoperators import BlockDiagonalOperator, MPI
from pyoperators.memory import empty
from pyoperators.utils import (
    ifirst, isscalarlike, product, split, strelapsed, strenum, strnbytes)
from . import _flib as flib
from .datatypes import Map, Tod
from .instruments import Instrument, Imager
from .packedtables import Layout, Sampling, Scene
from .mpiutils import gather_fitsheader_if_needed
from .operators import PointingMatrix, ProjectionOperator
from .wcsutils import (
    RotationBoresightEquatorialOperator, create_fitsheader, fitsheader2shape)
import numpy as np
import time

__all__ = ['Acquisition', 'AcquisitionImager', 'MaskPolicy']


class Acquisition(object):
    """
    The Acquisition class, which combines the instrument, sampling and scene
    models.

    """
    def __init__(self, instrument, sampling, scene, block=None,
                 max_nbytes=None, nprocs_instrument=None, nprocs_sampling=None,
                 comm=None):
        """
        Parameters
        ----------
        instrument : Instrument
            The Instrument instance.
        sampling : Sampling
            The sampling information (pointings, etc.)
        scene : Scene
            Discretization of the observed scene.
        block : tuple of slices, optional
            Partition of the samplings.
        max_nbytes : int or None, optional
            Maximum number of bytes to be allocated for the acquisition's
            operator.
        nprocs_instrument : int
            For a given sampling slice, number of procs dedicated to
            the instrument.
        nprocs_sampling : int
            For a given detector slice, number of procs dedicated to
            the sampling.
        comm : mpi4py.MPI.Comm
            The acquisition's MPI communicator. Note that it is transformed
            into a 2d cartesian communicator before being stored as the 'comm'
            attribute. The following relationship must hold:
                comm.size = nprocs_instrument * nprocs_sampling

        """
        if not isinstance(instrument, Instrument):
            raise TypeError(
                "The instrument input has an invalid type '{}'.".format(
                    type(instrument).__name__))
        if not isinstance(sampling, Sampling):
            raise TypeError(
                "The sampling input has an invalid type '{}'.".format(
                    type(instrument).__name__))
        if not isinstance(scene, Scene):
            raise TypeError(
                "The scene input has an invalid type '{}'.".format(
                    type(scene).__name__))

        if comm is None:
            comm = MPI.COMM_WORLD
        if nprocs_instrument is None and nprocs_sampling is None:
            nprocs_sampling = comm.size
        if nprocs_instrument is None:
            if nprocs_sampling < 1 or nprocs_sampling > comm.size:
                raise ValueError("Invalid value for nprocs_sampling '{0}'.".
                                 format(nprocs_sampling))
            nprocs_instrument = comm.size // nprocs_sampling
        elif nprocs_sampling is None:
            if nprocs_instrument < 1 or nprocs_sampling > comm.size:
                raise ValueError("Invalid value for nprocs_instrument '{0}'.".
                                 format(nprocs_instrument))
            nprocs_sampling = comm.size // nprocs_instrument
        if nprocs_instrument * nprocs_sampling != comm.size:
            raise ValueError('Invalid MPI distribution of the acquisition.')

        commgrid = comm.Create_cart(
            [nprocs_sampling, nprocs_instrument], reorder=True)

        comm_instrument = commgrid.Sub([False, True])
        comm_sampling = commgrid.Sub([True, False])

        self.scene = scene
        self.instrument = instrument.scatter(comm_instrument)
        self.sampling = sampling.scatter(comm_sampling)
        self.comm = commgrid
        if block is None:
            if max_nbytes is None:
                block = slice(0, len(self.sampling)),
            else:
                nbytes = self.get_operator_nbytes()
                if nbytes <= max_nbytes:
                    block = slice(0, len(self.sampling)),
                else:
                    nblocks = int(np.ceil(nbytes / max_nbytes))
                    block = split(len(self.sampling), nblocks)
        elif not isinstance(block, (list, tuple)) or \
             any(not isinstance(b, slice) for b in block):
            raise TypeError("Invalid block argument '{}'.".format(block))
        self.block = block

    def __getitem__(self, x):
        """
        Used to select instrument detectors and samplings.

        new_acq = acq[selection_instrument, selection_sampling]

        """
        out = copy(self)
        if isinstance(x, tuple):
            if len(x) == 0:
                x = (Ellipsis, Ellipsis)
            elif len(x) == 1:
                x += (Ellipsis,)
            elif len(x) > 2:
                raise ValueError('Invalid selection.')
            instrument = self.instrument[x[0]]
            sampling = self.sampling[x[1]]
            #XXX FIX BLOCKS!!!
        else:
            instrument = self.instrument[x]
            sampling = self.sampling
        out.instrument = instrument
        out.sampling = sampling
        return out

    def __str__(self):
        return '{}\nSamplings: {}'.format(self.instrument, len(self.sampling))

    __repr__ = __str__

    def pack(self, x, out=None, copy=False):
        return self.instrument.detector.pack(x)
    pack.__doc__ = Layout.pack.__doc__

    def unpack(self, x, out=None, missing_value=None, copy=False):
        return self.instrument.detector.unpack(x)
    unpack.__doc__ = Layout.unpack.__doc__

    def get_invntt_operator(self):
        """
        Return the inverse time-time noise correlation matrix as an Operator.

        """
        return BlockDiagonalOperator(
            [self.instrument.get_invntt_operator(self.sampling[b])
             for b in self.block], axisin=1)

    def get_projection_operator(self, header, npixels_per_sample=0,
                                method=None, units=None, derived_units=None,
                                verbose=True):
        if method == 'nearest':
            npixels_per_sample = 1
        time0 = time.time()
        info = []
        matrix = self.get_projection_matrix(
            header, npixels_per_sample=npixels_per_sample, method=method)
        npps_min = matrix.header['min_npixels_per_sample']

        if npps_min == 0:
            info += ['Warning, all detectors fall outside the map.']

        elif npixels_per_sample < npps_min:
            if npixels_per_sample > 0:
                info += ["Warning, the value 'npixels_per_sample' is not large"
                         " enough. Set it to '{0}' instead of '{1}' to avoid r"
                         "ecomputation of the pointing matrix.".format(
                             npps_min, npixels_per_sample)]
            else:
                info += ["Set the keyword 'npixels_per_sample' to '{0}' to avo"
                         "id recomputation of the pointing matrix.".format(
                             npps_min)]
            npixels_per_sample = npps_min
            matrix = self.get_projection_matrix(
                header, npixels_per_sample=npixels_per_sample, method=method)

        elif npixels_per_sample > npps_min:
            info += ["Warning, the value 'npixels_per_sample' is too large. Se"
                     "t it to '{0}' instead of '{1}' for better memory perform"
                     "ance.".format(npps_min, npixels_per_sample)]

        if matrix.header['outside']:
            info += ['Warning, some detectors fall outside the map.']

        info.insert(0, strnbytes(matrix.nbytes))
        print(strelapsed(time0, 'Computing the projector') + ' ({0})'.
              format(' '.join(info)))

        return ProjectionOperator(matrix, attrin={'_header': header},
                                  classin=Map, classout=Tod, units=units,
                                  derived_units=derived_units)

    def get_projection_nbytes(self, **keywords):
        """
        Return the number of bytes of the projection matrix.

        """
        keywords['verbose'] = False
        acq = type(self)(self.instrument, self.sampling[0], self.scene)
        proj = acq.get_projection_operator(**keywords)
        return proj.nbytes * len(self.sampling)

    def get_noise(self, out=None):
        """
        Return the noise realization according the instrument's noise model.

        Parameters
        ----------
        out : ndarray, optional
            Placeholder for the output noise.

        """
        if out is None:
            out = empty((len(self.instrument), len(self.sampling)))
        for b in self.block:
            self.instrument.get_noise(self.sampling[b], out=out[:, b])
        return out

    @classmethod
    def create_scan(cls, center, length, step, period, speed, acceleration,
                    nlegs=3, angle=0, instrument_angle=45, cross_scan=True,
                    dtype=None):
        """
        Return a sky scan as a Pointing array.

        Parameters
        ----------
        center : tuple
            (Right Ascension, declination) of the scan center.
        length : float
            Length of the scan lines, in arcseconds.
        step : float
            Separation between scan legs, in arcseconds.
        period : float
            Duration between two pointings.
        speed : float
            Scan speed, in arcsec/s.
        acceleration : float
            Acceleration, in arcsec/s^2.
        nlegs : integer
            Number of scan legs.
        angle : float
            Angle between the scan line direction and the North minus
            90 degrees, in degrees.
        instrument_angle : float
            Angle between the scan line direction and the instrument second
            axis, in degrees.
        cross_scan : boolean
            If true, a cross-scan is appended to the pointings.
        dtype : structured dtype
            Pointing data type.

        """
        raise NotImplementedError()
        scan = _create_scan(center, length, step, period, speed, acceleration,
                            nlegs, angle, dtype)
        if cross_scan:
            cross = _create_scan(center, length, step, period, speed,
                                 acceleration, nlegs, angle + 90, dtype)
            cross.time += scan.time[-1] + period
            scan, scan.header = (np.hstack([scan, cross]).view(),
                                 scan.header)

        scan.pa = angle + instrument_angle
        scan.header['HIERARCH instrument_angle'] = instrument_angle
        return scan

    def plot(self, map=None, header=None, new_figure=True, percentile=0.01,
             **keywords):
        """
        map : ndarray of dim 2
            The optional map to be displayed as background.
        header : pyfits.Header
            The optional map's FITS header.
        new_figure : boolean
            If true, plot the scan in a new window.
        percentile : float, tuple of two floats
            As a float, percentile of values to be discarded, otherwise,
            percentile of the minimum and maximum values to be displayed.

        """
        if header is None:
            header = getattr(map, 'header', None)
            if header is None:
                header = self.sampling.get_map_header(naxis=1)
        annim = self.sampling.plot(
            map=map, header=header, new_figure=new_figure,
            percentile=percentile, **keywords)
        return annim


class AcquisitionImager(Acquisition):
    """
    The AcquisitionImager class, which represents the setups of an imager
    instrument and the pointing.

    """
    def __init__(self, instrument, sampling, scene):
        """
        Parameters
        ----------
        instrument : Imager
            The Imager instance.
        sampling : Sampling
            The sampling information (pointings, etc.)
        scene : Scene
            Discretization of the observed scene.

        """
        if not isinstance(instrument, Imager):
            raise TypeError("The instrument has an invalid type '{}'.".format(
                            type(instrument).__name__))
        Acquisition.__init__(self, instrument, sampling, scene)
        self.object2world = RotationBoresightEquatorialOperator(self.sampling)

    def get_projection_matrix(self, header, npixels_per_sample=0, method=None):
        """
        Return the pointing matrix for a given set of pointings.

        Parameters
        ----------
        header : pyfits.Header
            The map FITS header
        npixels_per_sample : int
            Maximum number of sky pixels intercepted by a detector.
            By setting 0 (the default), the actual value will be determined
            automatically.
        method : string
            'sharp' : the intersection of the sky pixels and the detectors
                      is computed assuming that the transmission outside
                      the detector is zero and one otherwise (sharp edge
                      geometry)
            'nearest' : the value of the sky pixel closest to the detector
                        center is taken as the sample value, assuming
                        surface brightness conservation.

        """
        if method is None:
            if hasattr(self.instrument.detector, 'vertex'):
                method = 'sharp'
            else:
                method = 'nearest'
        method = method.lower()
        choices = ('nearest', 'sharp')
        if method not in choices:
            raise ValueError("Invalid method '" + method + "'. Expected values"
                             " are " + strenum(choices) + '.')

        comm = self.sky.comm
        header = gather_fitsheader_if_needed(header, comm=comm)
        shape_input = fitsheader2shape(header)
        if product(shape_input) > np.iinfo(np.int32).max:
            raise RuntimeError('The map is too large: pixel indices cannot be '
                               'stored using 32 bits: {0}>{1}'.format(product(
                               shape_input), np.iinfo(np.int32).max))

        if method == 'nearest':
            npixels_per_sample = 1

        # allocate memory for the pointing matrix
        ndetectors = len(self.instrument)
        nsamples = len(self.sampling)
        shape = (ndetectors, nsamples, npixels_per_sample)
        matrix = PointingMatrix.empty(shape, shape_input, verbose=False)

        # compute the pointing matrix
        if method == 'sharp':
            coords = self.instrument.image2object(
                self.instrument.detector.vertex)
            new_npps, outside = self._object2pmatrix_sharp_edges(
                coords, self.sampling, header, matrix, npixels_per_sample)
        elif method == 'nearest':
            coords = self.instrument.image2object(
                self.instrument.detector.center)
            new_npps = 1
            outside = self._object2pmatrix_nearest_neighbour(
                coords, self.sampling, header, matrix)
        else:
            raise NotImplementedError()

        matrix.header['method'] = method
        matrix.header['outside'] = bool(outside)
        matrix.header['HIERARCH min_npixels_per_sample'] = new_npps

        return matrix

    def plot(self, map=None, header=None, instrument=True, new_figure=True,
             percentile=0.01, **keywords):
        """
        map : ndarray of dim 2
            The optional map to be displayed as background.
        header : pyfits.Header
            The optional map's FITS header.
        instrument : boolean
            If true, plot the instrument's footprint on the sky.
        new_figure : boolean
            If true, plot the scan in a new window.
        percentile : float, tuple of two floats
            As a float, percentile of values to be discarded, otherwise,
            percentile of the minimum and maximum values to be displayed.

        """
        image = Acquisition.plot(
            self, map=map, header=header, new_figure=new_figure,
            percentile=percentile, **keywords)
        if not instrument:
            return image

        if hasattr(self.sampling, 'masked'):
            p = self.sampling[ifirst(self.sampling.masked, False)]
        else:
            p = self.sampling[0]
        t = RotationBoresightEquatorialOperator(p) * \
            self.instrument.image2object
        f = lambda x: image.projection.topixel(t(x))
        self.instrument.plot(transform=f, autoscale=False)
        return image

    @staticmethod
    def _object2ad(coords, pointing):
        """
        Convert coordinates in the instrument frame into celestial coordinates,
        assuming a pointing direction and a position angle.

        The instrument frame is the frame used for the 'center' or 'corner'
        coordinates, which are fields of the 'detector' attribute.

        Parameters
        ----------
        coords : float array (last dimension is 2)
            Coordinates in the instrument frame in arc seconds.
        pointing : array with flexible dtype including 'ra', 'dec' and 'pa'
            Direction corresponding to (0,0) in the local frame, in degrees.

        Returns
        -------
        coords_converted : float array (last dimension is 2, for Ra and Dec)
            Converted coordinates, in degrees.

        Notes
        -----
        The routine is not accurate at the poles.

        """
        coords = np.array(coords, float, order='c', copy=False)
        shape = coords.shape
        coords = coords.reshape((-1, 2))
        new_shape = (pointing.size,) + coords.shape
        result = np.empty(new_shape, float)
        for r, ra, dec, pa in zip(result, pointing['ra'].flat,
                                  pointing['dec'].flat, pointing['pa'].flat):
            flib.wcsutils.object2ad(coords.T, r.T, ra, dec, pa)
        result = result.reshape(pointing.shape + shape)
        return result

    @staticmethod
    def _object2xy_minmax(coords, pointing, header):
        """
        Return the minimum and maximum sky pixel coordinate values given a set
        of coordinates specified in the instrument frame.

        """
        coords = np.array(coords, float, order='c', copy=False)
        xmin, ymin, xmax, ymax, status = flib.wcsutils.object2xy_minmax(
            coords.reshape((-1, 2)).T, pointing['ra'].ravel(),
            pointing['dec'].ravel(), pointing['pa'].ravel(),
            str(header).replace('\n', ''))
        if status != 0:
            raise RuntimeError()
        return xmin, ymin, xmax, ymax

    @staticmethod
    def _object2pmatrix_sharp_edges(coords, pointing, header, pmatrix,
                                    npixels_per_sample):
        """
        Return the sparse pointing matrix whose values are intersections
        between detectors and map pixels.

        """
        coords = coords.reshape((-1,) + coords.shape[-2:])
        ra = pointing['ra'].ravel()
        dec = pointing['dec'].ravel()
        pa = pointing['pa'].ravel()
        masked = pointing['masked'].view(np.int8).ravel()
        if pmatrix.size == 0:
            # f2py doesn't accept zero-sized opaque arguments
            pmatrix = np.empty(1, np.int64)
        else:
            pmatrix = pmatrix.ravel().view(np.int64)
        header = str(header).replace('\n', '')

        new_npps, out, status = flib.wcsutils.object2pmatrix_sharp_edges(
            coords.T, ra, dec, pa, masked, header, pmatrix, npixels_per_sample)
        if status != 0:
            raise RuntimeError()

        return new_npps, out

    @staticmethod
    def _object2pmatrix_nearest_neighbour(coords, pointing, header,
                                          pmatrix):
        """
        Return the sparse pointing matrix whose values are intersection between
        detector centers and map pixels.

        """
        coords = coords.reshape((-1, 2))
        area = np.ones(coords.shape[0])
        ra = pointing['ra'].ravel()
        dec = pointing['dec'].ravel()
        pa = pointing['pa'].ravel()
        masked = pointing['masked'].view(np.int8).ravel()
        pmatrix = pmatrix.ravel().view(np.int64)
        header = str(header).replace('\n', '')
        out, status = flib.wcsutils.object2pmatrix_nearest_neighbour(
            coords.T, area, ra, dec, pa, masked, header, pmatrix)
        if status != 0:
            raise RuntimeError()

        return out


def _create_scan(center, length, step, period, speed, acceleration, nlegs,
                 angle, dtype):
    """
    compute the pointing timeline of the instrument reference point
    from the description of a scan map
    Authors: R. Gastaud, P. Chanial

    """
    raise NotImplementedError()
    ra0, dec0 = center

    length = float(length)
    if length <= 0:
        raise ValueError('Input length must be strictly positive.')

    step = float(step)
    if step <= 0:
        raise ValueError('Input step must be strictly positive.')

    period = float(period)
    if period <= 0:
        raise ValueError('Input sampling period must be strictly positive.')

    speed = float(speed)
    if speed <= 0:
        raise ValueError('Input speed must be strictly positive.')

    acceleration = float(acceleration)
    if acceleration <= 0:
        raise ValueError('Input acceleration must be strictly positive.')

    nlegs = int(nlegs)
    if nlegs <= 0:
        raise ValueError('Input nlegs must be strictly positive.')

    # compute the different times and the total number of points
    # acceleration time at the beginning of a leg, and deceleration time
    # at the end
    # The time needed to turn around is 2 * (extra_time1 + extra_time2)
    extra_time1 = speed / acceleration
    # corresponding length
    extralength = 0.5 * acceleration * extra_time1 * extra_time1
    # Time needed to go from a scan line to the next
    extra_time2 = np.sqrt(step / acceleration)

    # Time needed to go along the scanline at constant speed
    line_time = length / speed
    # Total time for a scanline
    full_line_time = extra_time1 + line_time + extra_time1 + 2 * extra_time2
    # Total duration of the observation
    total_time = full_line_time * nlegs - 2 * extra_time2

    # Number of samples
    nsamples = int(np.ceil(total_time / period))

    # initialization
    time          = np.zeros(nsamples)
    latitude      = np.zeros(nsamples)
    longitude     = np.zeros(nsamples)
    infos         = np.zeros(nsamples, dtype=int)
    line_counters = np.zeros(nsamples, dtype=int)

    # Start of computations, alpha and delta are the longitude and
    # latitide in arc seconds in the referential of the map.
    sign = 1
    delta = -extralength - length/2
    alpha = -step * (nlegs-1)/2
    alpha0 = alpha
    line_counter = 0
    working_time = 0.

    for i in range(nsamples):
        info = 0

        # check if new line
        if working_time > full_line_time:
            working_time = working_time - full_line_time
            sign = -sign
            line_counter = line_counter + 1
            alpha = -step * (nlegs-1) / 2 + line_counter * step
            alpha0 = alpha

        # acceleration at the beginning of a scan line to go from 0 to the
        # speed.
        if working_time < extra_time1:
            delta = -sign * (extralength + length / 2) + sign * 0.5 * \
                    acceleration * working_time * working_time
            info = Pointing.TURNAROUND

        # constant speed
        if working_time >= extra_time1 and \
           working_time < extra_time1 + line_time:
            delta = sign * (-length / 2 + (working_time - extra_time1) * speed)
            info = Pointing.INSCAN

        # Deceleration at then end of the scanline to stop
        if working_time >= extra_time1 + line_time and \
           working_time < extra_time1 + line_time + extra_time1:
            dt = working_time - extra_time1 - line_time
            delta = sign * (length / 2 + speed * dt -
                    0.5 * acceleration * dt**2)
            info = Pointing.TURNAROUND

        # Acceleration to go toward the next scan line
        if working_time >= 2 * extra_time1 + line_time and \
           working_time < 2 * extra_time1 + line_time + extra_time2:
            dt = working_time - 2 * extra_time1 - line_time
            alpha = alpha0 + 0.5 * acceleration * dt**2
            info = Pointing.TURNAROUND

        # Deceleration to stop at the next scan line
        if working_time >= 2 * extra_time1 + line_time + extra_time2 and \
           working_time < full_line_time:
            dt = working_time - 2 * extra_time1 - line_time - extra_time2
            alpha = (alpha0 + step / 2) + acceleration * extra_time2 * dt - \
                    0.5 * acceleration * dt**2
            info = Pointing.TURNAROUND

        time[i] = i * period
        infos[i] = info
        latitude[i] = delta
        longitude[i] = alpha
        line_counters[i] = line_counter
        working_time = working_time + period

    # Convert the longitude and latitude *expressed in degrees) to ra and dec
    ra, dec = _change_coord(ra0, dec0, angle - 90, longitude / 3600,
                            latitude / 3600)

    scan = Pointing((ra, dec, 0.), time, infos, dtype=dtype)
    header = create_fitsheader(nsamples)
    header['ra'] = ra0
    header['dec'] = dec0
    header['HIERARCH scan_angle'] = angle
    header['HIERARCH scan_length'] = length
    header['HIERARCH scan_nlegs'] = nlegs
    header['HIERARCH scan_step'] = step
    header['HIERARCH scan_speed'] = speed
    header['HIERARCH scan_acceleration'] = acceleration
    scan.header = header
    return scan


def _change_coord(ra0, dec0, pa0, lon, lat):
    """
    Transforms the longitude and latitude coordinates expressed in the
    native coordinate system attached to a map into right ascension and
    declination.
    Author: R. Gastaud

    """
    from numpy import arcsin, arctan2, cos, deg2rad, mod, sin, rad2deg

    # Arguments in radian
    lambd = deg2rad(lon)
    beta  = deg2rad(lat)
    alpha = deg2rad(ra0)
    delta = deg2rad(dec0)
    eps   = deg2rad(pa0)

    # Cartesian coordinates from longitude and latitude
    z4 = sin(beta)
    y4 = sin(lambd)*cos(beta)
    x4 = cos(lambd)*cos(beta)

    # rotation about the x4 axe of -eps
    x3 =  x4
    y3 =  y4*cos(eps) + z4*sin(eps)
    z3 = -y4*sin(eps) + z4*cos(eps)

    # rotation about the axis Oy2, angle delta
    x2 = x3*cos(delta) - z3*sin(delta)
    y2 = y3
    z2 = x3*sin(delta) + z3*cos(delta)

    # rotation about the axis Oz1, angle alpha
    x1 = x2*cos(alpha) - y2*sin(alpha)
    y1 = x2*sin(alpha) + y2*cos(alpha)
    z1 = z2

    # Compute angles from cartesian coordinates
    # it is the only place where we can get nan with arcsinus
    dec = rad2deg(arcsin(np.clip(z1, -1., 1.)))
    ra  = mod(rad2deg(arctan2(y1, x1)), 360.)

    return ra, dec


class MaskPolicy(object):
    KEEP   = 0
    MASK   = 1
    REMOVE = 2

    def __init__(self, flags, values, description=None):
        self._description = description
        if isscalarlike(flags):
            if isinstance(flags, str):
                flags = flags.split(',')
            else:
                flags = [flags]
        flags = [_.strip() for _ in flags]
        if isscalarlike(values):
            if isinstance(values, str):
                values = values.split(',')
            else:
                values = [values]
        values = [_.strip().lower() for _ in values]
        if len(flags) != len(values):
            raise ValueError('The number of policy flags is different from the'
                             ' number of policies.')

        self._policy = []
        for flag, value in zip(flags, values):
            if flag[0] == '_':
                raise ValueError('A policy flag should not start with an under'
                                 'score.')
            choices = 'keep', 'mask', 'remove'
            if value not in choices:
                raise KeyError('Invalid policy {}={!r}. Expected values are {}'
                               '.'.format(flag, value, strenum(choices)))
            setattr(self, flag, value)
        self._flags = flags

    def __array__(self, dtype=int):
        conversion = {'keep': self.KEEP,
                      'mask': self.MASK,
                      'remove': self.REMOVE}
        return np.array([conversion[getattr(self, _)]
                         for _ in self._flags], dtype=dtype)

    def __str__(self):
        s = self._description + ': ' if self._description is not None else ''
        s += ', '.join('{}={!r}'.format(_, getattr(self, _))
                       for _ in self._flags)
        return s



import numpy as np

from numpy.testing import assert_allclose, assert_array_equal
from pyoperators.utils.testing import assert_eq
from pysimulators.datautils import (
    airy_disk, distance, gaussian, profile, integrated_profile, psd2,
    _distance_slow)


def test_distance1():
    origin = (0.,)
    d0 = np.arange(5.)*0.5
    d1 = distance(5, origin=origin, resolution=0.5)
    d2 = _distance_slow((5,), origin, [0.5], None)
    assert_array_equal(d0, d1)
    assert_array_equal(d0, d2)


def test_distance2():
    origin = (0., 1.)
    d0 = np.array([[2., np.sqrt(5), np.sqrt(8)], [0, 1, 2]])
    d1 = distance((2, 3), origin=origin, resolution=(1., 2.))
    d2 = _distance_slow((2, 3), origin, [1., 2.], None)
    assert_array_equal(d0, d1)
    assert_array_equal(d0, d2)


def test_fwhm():
    def func(f, fwhm, resolution, n):
        m = f((1000, 1000), fwhm=fwhm, resolution=resolution)
        assert np.sum(m[500, :] > np.max(m)/2) == n
    for f in [gaussian, airy_disk]:
        for fwhm, resolution, n in zip([10, 10, 100], [0.1, 1, 10],
                                       [100, 10, 10]):
            yield func, f, fwhm, resolution, n


def test_profile():
    def profile_slow(input, origin=None, bin=1.):
        d = distance(input.shape, origin=origin)
        d /= bin
        d = d.astype(int)
        m = int(np.max(d))
        p = np.ndarray(int(m+1))
        n = np.zeros(m+1, int)
        for i in range(m+1):
            p[i] = np.mean(input[d == i])
            n[i] = np.sum(d == i)
        return p, n
    d = distance((10, 20))
    x, y, n = profile(d, origin=(4, 5), bin=2., histogram=True)
    y2, n2 = profile_slow(d, origin=(4, 5), bin=2.)
    assert_eq(y, y2[0:y.size])
    assert_eq(n, n2[0:n.size])


def test_integrated_profile():
    def integrated_profile_slow(input, origin=None, bin=1.):
        d = distance(input.shape, origin=origin)
        m = int(np.max(d))
        x = np.ndarray(int(m+1))
        y = np.ndarray(int(m+1))
        for i in range(m+1):
            x[i] = (i + 1) * bin
            y[i] = np.sum(input[d < x[i]])
        return x, y
    d = distance((10, 20))
    x, y = integrated_profile(d, origin=(4, 5), bin=2.)
    x2, y2 = integrated_profile_slow(d, origin=(4, 5), bin=2.)
    assert_eq(x, x2[0:y.size])
    assert_eq(y, y2[0:y.size])


def test_psd2():
    image = np.random.randn(122, 122)
    fs = 1 / 0.11
    psd = psd2(image, sampling_frequency=fs)
    assert_allclose(np.sum(psd*(fs**2/image.size)), np.mean(image**2))



from __future__ import division

import healpy as hp
import itertools
import numpy as np
from numpy.testing import assert_allclose, assert_equal
from pyoperators import IdentityOperator, CompositionOperator
from pyoperators.utils.testing import (
    assert_is_type, assert_raises, assert_same)
from pysimulators.interfaces.healpy import (
    Cartesian2HealpixOperator, Healpix2CartesianOperator,
    Spherical2HealpixOperator, Healpix2SphericalOperator,
    HealpixConvolutionGaussianOperator, HealpixLaplacianOperator)

NSIDE = 512


def test_cartesian_healpix():
    vecs = ((1, 0, 0), (0, 1, 0), (0, 0, 1),
            ((1, 0, 0), (0, 1, 0), (0, 0, 1)),
            (((1, 0, 0), (0, 1, 0)),))
    shapes = ((), (), (), (3,), (1, 2))

    def func(n, v, s):
        c2h = Cartesian2HealpixOperator(NSIDE, nest=n)
        h2c = Healpix2CartesianOperator(NSIDE, nest=n)
        a = h2c(c2h(v))
        assert_equal(a.shape, s + (3,))
        assert_allclose(a, v, atol=1e-1)
    for n in False, True:
        for v, s in zip(vecs, shapes):
            yield func, n, v, s


def test_cartesian_healpix_rules():
    op = Cartesian2HealpixOperator(NSIDE)
    assert_is_type(op.I, Healpix2CartesianOperator)
    assert_equal(op.I.nside, op.nside)
    assert_is_type(Cartesian2HealpixOperator(NSIDE)(
        Healpix2CartesianOperator(2 * NSIDE)), CompositionOperator)
    assert_is_type(Cartesian2HealpixOperator(NSIDE)(
        Healpix2CartesianOperator(NSIDE)), IdentityOperator)


def test_cartesian_healpix_error():
    op = Cartesian2HealpixOperator(NSIDE)

    def func(i, o):
        if i.shape == (3,) and o.shape == ():
            op(i, o)
            return
        assert_raises(ValueError, op.__call__, i, o)
    for i, o in itertools.product((np.ones(()), np.ones(2), np.ones(3)),
                                  (np.empty(()), np.empty(2), np.empty(3))):
        yield func, i, o


def test_healpix_cartesian():
    vecs = (0, 100, 12*NSIDE**2-1,
            (0, 100, 12*NSIDE**2-1),
            ((0, 100),))
    shapes = ((), (), (), (3,), (1, 2))

    def func(n, v, s):
        h2c = Healpix2CartesianOperator(NSIDE, nest=n)
        c2h = Cartesian2HealpixOperator(NSIDE, nest=n)
        a = c2h(h2c(v))
        assert_equal(a.shape, s)
        assert_equal(a, v)
    for n in False, True:
        for v, s in zip(vecs, shapes):
            yield func, n, v, s


def test_healpix_cartesian_rules():
    op = Healpix2CartesianOperator(NSIDE)
    assert_is_type(op.I, Cartesian2HealpixOperator)
    assert_equal(op.I.nside, op.nside)
    assert_is_type(Healpix2CartesianOperator(NSIDE)(
        Cartesian2HealpixOperator(2 * NSIDE)), CompositionOperator)
    assert_is_type(Healpix2CartesianOperator(NSIDE)(
        Cartesian2HealpixOperator(NSIDE)), IdentityOperator)


def test_healpix_cartesian_error():
    op = Healpix2CartesianOperator(NSIDE)

    def func(i, o):
        if i.shape == () and o.shape == (3,):
            op(i, o)
            return
        assert_raises(ValueError, op.__call__, i, o)
    for i, o in itertools.product((np.array(1.), np.zeros(2), np.zeros(3)),
                                  (np.array(1.), np.zeros(2), np.zeros(3))):
        yield func, i, o


def test_spherical_healpix():
    dirs_za = ((10, 0), (20, 0), (130, 0), (10, 20), (20, 130),
               ((10, 0), (20, 0), (130, 0), (10, 20), (20, 130)),
               (((10, 0), (20, 200), (130, 300)),))
    dirs_az = ((0, 10), (0, 20), (0, 130), (20, 10), (130, 20),
               ((0, 10), (0, 20), (0, 130), (20, 10), (130, 20)),
               (((0, 10), (200, 20), (300, 130)),))
    dirs_ea = ((80, 0), (70, 0), (-40, 0), (80, 20), (70, 130),
               ((80, 0), (70, 0), (-40, 0), (80, 20), (70, 130)),
               (((80, 0), (70, 200), (-40, 300)),))
    dirs_ae = ((0, 80), (0, 70), (0, -40), (20, 80), (130, 70),
               ((0, 80), (0, 70), (0, -40), (20, 80), (130, 70)),
               (((0, 80), (200, 70), (300, -40)),))
    shapes = ((), (), (), (), (), (5,), (1, 3))

    op_ref = Spherical2HealpixOperator(NSIDE, 'zenith,azimuth')
    refs = [op_ref(np.radians(v)) for v in dirs_za]

    def func(c, v, s, r):
        s2h = Spherical2HealpixOperator(NSIDE, c)
        h2s = Healpix2SphericalOperator(NSIDE, c)
        assert_allclose(s2h(np.radians(v)), r)
        a = np.degrees(h2s(s2h(np.radians(v))))
        assert_equal(a.shape, s + (2,))
        assert_allclose(a, v, rtol=1e-2, atol=0.5)
    for c, d in (('zenith,azimuth', dirs_za),
                 ('azimuth,zenith', dirs_az),
                 ('elevation,azimuth', dirs_ea),
                 ('azimuth,elevation', dirs_ae)):
        for v, s, r in zip(d, shapes, refs):
            yield func, c, v, s, r


def test_spherical_healpix_rules():
    def func(op, nside, c, nest):
        op2 = Healpix2SphericalOperator(nside, c, nest=nest)
        if c == op.convention and nside == op.nside and nest == op.nest:
            assert_is_type(op(op2), IdentityOperator)
        else:
            assert_is_type(op(op2), CompositionOperator)
    for c in Spherical2HealpixOperator.CONVENTIONS:
        op = Spherical2HealpixOperator(NSIDE, c, nest=True)
        assert_is_type(op.I, Healpix2SphericalOperator)
        assert_equal(op.I.convention, op.convention)
        for nside in NSIDE, 2*NSIDE:
            for c2 in Spherical2HealpixOperator.CONVENTIONS:
                for nest in False, True:
                    yield func, op, nside, c2, nest


def test_spherical_healpix_error():
    assert_raises(ValueError, Spherical2HealpixOperator, NSIDE, 'bla')
    op = Spherical2HealpixOperator(NSIDE, 'zenith,azimuth')

    def func(i, o):
        if i.shape == (2,) and o.shape == ():
            op(i, o)
            return
        assert_raises(ValueError, op.__call__, i, o)
    for i, o in itertools.product((np.array(1.), np.zeros(2), np.zeros(3)),
                                  (np.array(1.), np.zeros(2), np.zeros(3))):
        yield func, i, o


def test_healpix_spherical():
    vecs = (0, 100, 12*NSIDE**2-1,
            (0, 100, 12*NSIDE**2-1),
            ((0, 100),))
    shapes = ((), (), (), (3,), (1, 2))

    def func(c, n, v, s):
        h2s = Healpix2SphericalOperator(NSIDE, c, nest=n)
        s2h = Spherical2HealpixOperator(NSIDE, c, nest=n)
        a = s2h(h2s(v))
        assert_equal(a.shape, s)
        assert_equal(a, v)
    for c in Healpix2SphericalOperator.CONVENTIONS:
        for n in False, True:
            for v, s in zip(vecs, shapes):
                yield func, c, n, v, s


def test_healpix_spherical_rules():
    def func(op, nside, c, nest):
        op2 = Spherical2HealpixOperator(nside, c, nest=nest)
        if c == op.convention and nside == op.nside and nest == op.nest:
            assert_is_type(op(op2), IdentityOperator)
        else:
            assert_is_type(op(op2), CompositionOperator)
    for c in Healpix2SphericalOperator.CONVENTIONS:
        op = Healpix2SphericalOperator(NSIDE, c, nest=True)
        assert_is_type(op.I, Spherical2HealpixOperator)
        assert_equal(op.I.convention, op.convention)
        for nside in NSIDE, 2*NSIDE:
            for c2 in Healpix2SphericalOperator.CONVENTIONS:
                for nest in False, True:
                    yield func, op, nside, c2, nest


def test_healpix_spherical_error():
    assert_raises(ValueError, Healpix2SphericalOperator, NSIDE, 'bla')
    op = Healpix2SphericalOperator(NSIDE, 'zenith,azimuth')

    def func(i, o):
        if i.shape == () and o.shape == (2,):
            op(i, o)
            return
        assert_raises(ValueError, op.__call__, i, o)
    for i, o in itertools.product((np.array(1.), np.zeros(2), np.zeros(3)),
                                  (np.array(1.), np.zeros(2), np.zeros(3))):
        yield func, i, o


def test_healpix_convolution():
    nside = 16
    input = np.arange(12 * nside**2)
    keywords = {'fwhm': np.radians(30),
                'iter': 2,
                'lmax': 8,
                'use_weights': True}
    expected = hp.smoothing(input, **keywords)

    op = HealpixConvolutionGaussianOperator(**keywords)
    assert_same(op(input), expected)

    input = np.array([input, input]).T
    assert_same(op(input), np.array([expected, expected]).T)


def test_healpix_laplacian():
    nside = 1
    ipix = 4
    npix = 12
    map = np.zeros(npix)
    map[ipix] = 1
    L = HealpixLaplacianOperator(nside)
    assert L.flags.square
    assert L.flags.symmetric
    h2 = np.array(4 * np.pi / npix)
    expected = [1, 0, 0, 1, -20, 4, 0, 4, 1, 0, 0, 1] / (6 * h2)
    assert_same(L(map), expected)
    assert_same(L(np.repeat(map, 5).reshape(12, 5)),
                expected[:, None], broadcasting=True)



import numpy as np
import glob
import os
import pickle

import astropy.io.fits as fits
from numpy.testing import assert_equal
from pyoperators.utils.testing import assert_eq, assert_is_none
from pysimulators import Quantity, FitsArray, Map, Tod, create_fitsheader
from uuid import uuid1

filename = 'pysimulatorstest-'+str(uuid1())
types = (Quantity, FitsArray, Map, Tod)
keywords_q = {'unit': 'km', 'derived_units': {'km': Quantity(1000, 'm')}}
keywords_f = keywords_q.copy()
keywords_f.update({'header': fits.Header({'NAXIS': 0})})
keywords_m = keywords_f.copy()
keywords_m.update({'coverage': 8, 'error': 1., 'origin': 'upper'})
keywords_t = keywords_f.copy()
keywords_t.update({'mask': True})
keywords_types = (keywords_q, keywords_f, keywords_m, keywords_t)
dtypes = (bool, int, np.float32, np.float64, np.complex64, np.complex128)


def teardown():
    for file in glob.glob(filename+'*'):
        os.remove(file)

a = np.ones((4, 3))
a[1, 2] = 4
q = Quantity(a, unit='myunit', derived_units={'myunit': Quantity(2., 'Jy')})
header = create_fitsheader(fromdata=q, cdelt=0.5, crval=(4., 8.))
header['BUNIT'] = 'myunit'
f = FitsArray(q, header=header)
m = Map(f, origin='upper', error=a*2, coverage=a*3)
mask = np.zeros((4, 3), np.bool8)
mask[0, 2] = True
t = Tod(f, mask=mask)
del mask


def test_copy_false_subok_true():
    def func(obj1, t):
        obj2 = t(obj1, copy=False, subok=True)
        if isinstance(obj, t):
            assert obj1 is obj2
        else:
            assert obj1 is not obj2
            assert_equal(obj1, obj2)
    for obj in [q, f, m, t]:
        for ty in types:
            yield func, obj, ty


def test_copy_false_subok_false():
    def func(obj1, t):
        obj2 = t(obj1, copy=False, subok=False)
        if type(obj1) is t:
            assert obj1 is obj2
        else:
            assert obj1 is not obj2
            assert_equal(obj1, obj2)
    for obj in [q, f, m, t]:
        for ty in types:
            yield func, obj, ty


def test_copy_true_subok_true():
    def func(obj1, ty):
        obj2 = ty(obj1, copy=True, subok=True)
        assert obj1 is not obj2
        assert_equal(obj1, obj2)
        if isinstance(obj1, ty):
            assert type(obj2) is type(obj1)
        else:
            assert type(obj2) is ty
    for obj1 in [q, f, m, t]:
        for ty in types:
            yield func, obj1, ty


def test_copy_true_subok_false():
    def func(obj1, ty):
        obj2 = ty(obj1, copy=True, subok=False)
        assert obj1 is not obj2
        assert_equal(obj1, obj2)
        assert type(obj2) is ty
    for obj1 in [q, f, m, t]:
        for ty in types:
            yield func, obj1, ty


def test_input1():
    def func(t, i):
        d = t(i)
        assert d.shape == (1,)
        assert d[0] == 2
    for t in types:
        for i in [(2,), [2], np.array([2])]:
            yield func, t, i


def test_input2():
    def func(t, i):
        d = t(i)
        assert d.shape == ()
        assert d[...] == 2
    for t in types:
        for i in [2, np.array(2)]:
            yield func, t, i


def test_input3():
    def func(t):
        d = t([])
        assert d.shape == (0,)
    for t in types:
        yield func, t


def test_view():
    array = np.ones((10, 32))
    aq = ['unit', 'derived_units']
    af = aq + ['header']
    attrs = [aq, af, af + ['coverage', 'error', 'origin'], af + ['mask']]

    def func(t, attr):
        d = array.view(t)
        for a in attr:
            assert hasattr(d, a)
    for t, attr in zip(types, attrs):
        yield func, t, attr


def test_operation():
    def func(t):
        a = t.ones(10)
        a2 = a + 1
        assert type(a2) is t
    for t in types[1:]:
        yield func, t


def test_dtype():
    def func1(t, d):
        a = t(np.array([10, 20], dtype=d))
        assert a.dtype == (d if d != int else float)

    def func2(t, d):
        a = t(np.array([10, 20], dtype=d), dtype=d)
        assert a.dtype == d

    def func3(t, d):
        a = t(np.ones((4, 3)), dtype=d, copy=False)
        assert a.dtype == d
    for t in types:
        for d in dtypes:
            yield func1, t, d
            yield func2, t, d
            yield func3, t, d


def test_empty_ones_zeros():
    def func(t, k, m, d):
        method = getattr(t, m)
        a = method((2, 3), dtype=d, **k)
        assert_eq(a.dtype, t.default_dtype if d is None else d)
        for k_, v in k.items():
            assert_eq(getattr(a, k_), v)

    for t, k in zip(types, keywords_types):
        for m in ('empty', 'ones', 'zeros'):
            for d in (None,) + dtypes:
                yield func, t, k, m, d


def test_map():
    class MAP(Map):
        def __init__(self, data):
            self.info = 'info'
    m = MAP(np.ones(3))
    assert sorted(m.__dict__) == ['_derived_units', '_header', '_unit',
                                  'coverage', 'error', 'info', 'origin']


def test_pickling():
    def func(v, o):
        o2 = pickle.loads(pickle.dumps(o, v))
        assert_eq(o, o2)
    for v in range(pickle.HIGHEST_PROTOCOL):
        for o in (q, f, m, t):
            yield func, v, o


def test_save():
    def func(o):
        o.save(filename+'_obj.fits')
        o2 = type(o)(filename+'_obj.fits')
        assert_eq(o, o2)
    for o in (f, m, t):
        yield func, o


def test_ndarray_funcs():
    data = [[1, 2, 3, 4], [5, 6, 7, 8]]
    mask = [[True, False, False, False], [False, True, True, False]]
    funcs = np.min, np.max, np.sum, np.mean, np.ptp, np.round, np.std, np.var
    axes = (None, 0, 1)

    def f(cls, func, axis, m):
        keywords_array = {'unit': 'u'}
        if cls is Tod:
            keywords_array['mask'] = m
        array = cls(data, **keywords_array)
        keywords_func = {'axis': axis} if func is not np.round else {}
        result = func(array, **keywords_func)
        if cls is Tod:
            ref = func(np.ma.MaskedArray(array.magnitude, mask=m),
                       **keywords_func)
            if not isinstance(ref, np.ndarray):
                ref = np.ma.MaskedArray(ref)
        else:
            ref = func(array.magnitude, **keywords_func)
        assert_equal(result.view(np.ndarray), ref)
        if np.isscalar(result):
            return
        if func is np.var:
            assert result._unit == {'u': 2}
        else:
            assert result._unit == {'u': 1}
        if cls is Map:
            assert_is_none(result.coverage)
            assert_is_none(result.error)
        elif cls is Tod:
            if result.mask is None:
                assert not ref.mask
            else:
                assert_eq(result.mask, ref.mask)

    for cls in (Quantity, FitsArray, Map, Tod):
        if cls is Tod:
            masks = None, mask
        else:
            masks = None,
        for func in funcs:
            for axis in (None,) if func is np.round else axes:
                for m in masks:
                    yield f, cls, func, axis, m


def test_astype():
    dtypes = np.int8, np.int32, np.int64, np.float32, np.float64

    def func(d, e):
        m = Map(np.array([1, 2, 3], dtype=d), coverage=np.array([0, 1, 0],
                dtype=d), error=np.array([2, 2, 2], dtype=d))
        m2 = m.astype(e)
        assert m2.dtype == e
        assert m2.coverage.dtype == e
        assert m2.error.dtype == e
    for d in dtypes:
        for e in dtypes:
            yield func, d, e



from __future__ import absolute_import, division, print_function
from pyoperators import FFTOperator
from pyoperators.utils import isscalarlike, product
from . import _flib as flib
from .datatypes import Map
import numpy as np
import scipy.fftpack
import scipy.signal
import scipy.special

__all__ = [
    'airy_disk',
    'aperture_circular',
    'distance',
    'ds9',
    'gaussian',
    'integrated_profile',
    'phasemask_fourquadrant',
    'profile',
    'profile_psd2',
    'psd2',
]


def airy_disk(shape, fwhm=None, r0=None, origin=None, resolution=1):
    """
    Return a two-dimensional map with an Airy pattern.

    Parameters
    ----------
    shape : tuple
        The map shape (ny, nx).
    fwhw : float
        The Full Width Half Maximum of the primary lobe.
    f0 : float
        The radius of the first dark ring, where the intensity is zero.
    origin : tuple of float
        The pattern center of the form (x0,y0)
    resolution : float
        The pixel scale.

    """
    if fwhm is None and r0 is None:
        raise ValueError('No scale parameter.')
    d = distance(shape, origin=origin, resolution=resolution)
    index = np.where(d == 0)
    d[index] = 1.e-30
    if fwhm is not None:
        d *= 1.61633 / (fwhm / 2)
    else:
        d *= 3.8317 / r0
    d = (2 * scipy.special.jn(1, d) / d)**2
    d /= np.sum(d)
    return d


def aperture_circular(shape, diameter, origin=None, resolution=1):
    """
    Return a two-dimensional map with circular mask.

    Parameters
    ----------
    shape : tuple
        The map shape (ny, nx).
    fwhw : float
        The Full Width Half Maximum of the primary lobe.
    origin : tuple of float
        The pattern center of the form (x0,y0)
    resolution : float
        The pixel scale.

    """
    array = distance(shape, origin=origin, resolution=resolution)
    m = array > diameter / 2
    array[m] = 0
    array[~m] = 1
    return array


def distance(shape, origin=None, resolution=1.):
    """
    Returns an array whose values are the distances to a given origin.

    Parameters
    ----------
    shape : tuple of integer
        dimensions of the output array. For a 2d array, the first integer
        is for the Y-axis and the second one for the X-axis.
    origin : array-like in the form of (x0, y0, ...), optional
        The coordinates of the point from which the distance is calculated,
        assuming a zero-based coordinate indexing. Default value is the array
        center.
    resolution : array-like in the form of (dx, dy, ...), optional
        Inter-pixel distance. Default is one. If resolution is a Quantity, its
        unit will be carried over to the returned distance array

    Example
    -------
    nx, ny = 3, 3
    print(distance((ny,nx)))
    [[ 1.41421356  1.          1.41421356]
     [ 1.          0.          1.        ]
     [ 1.41421356  1.          1.41421356]]

    """
    if isscalarlike(shape):
        shape = (shape,)
    else:
        shape = tuple(shape)
    shape = tuple(int(np.round(s)) for s in shape)
    rank = len(shape)

    if origin is None:
        origin = (np.array(shape[::-1], dtype=float) - 1) / 2
    else:
        origin = np.ascontiguousarray(origin, dtype=float)

    unit = getattr(resolution, '_unit', None)

    if isscalarlike(resolution):
        resolution = np.resize(resolution, rank)
    resolution = np.asanyarray(resolution, dtype=float)

    du = flib.datautils
    if rank == 1:
        d = du.distance_1d(shape[0], origin[0], resolution[0])
    elif rank == 2:
        d = du.distance_2d(shape[1], shape[0], origin, resolution).T
    elif rank == 3:
        d = du.distance_3d(shape[2], shape[1], shape[0], origin, resolution).T
    else:
        d = _distance_slow(shape, origin, resolution, float)

    return Map(d, copy=False, unit=unit)


def _distance_slow(shape, origin, resolution, dtype):
    """
    Returns an array whose values are the distances to a given origin.

    This routine is written using np.meshgrid routine. It is slower
    than the Fortran-based `distance` routine, but can handle any number
    of dimensions.

    Refer to `pysimulators.distance` for full documentation.
    """

    if dtype is None:
        dtype = float
    index = []
    for n, o, r in zip(reversed(shape), origin, resolution):
        index.append(slice(0, n))
    d = np.asarray(np.mgrid[index], dtype).T
    d -= np.asanyarray(origin)
    d *= resolution
    np.square(d, d)
    d = Map(np.sqrt(np.sum(d, axis=d.shape[-1])), dtype=dtype, copy=False)
    return d


class Ds9(object):
    """
    Helper around the ds9 package.

    Examples
    --------
    >>> ds9.open_in_new_window = False
    >>> d = ds9.current
    >>> d.set('scale linear')
    >>> mynewmap = gaussian((128, 128), sigma=32)
    >>> ds9(mynewmap, 'zoom to fit')

    """
    def __call__(self, array, xpamsg=None, origin=None, **keywords):
        if not isinstance(array, str):
            array = np.asanyarray(array)
            dtype = array.dtype
        else:
            dtype = None
        array = Map(array, dtype=dtype, copy=False, origin=origin)
        array.ds9(xpamsg=xpamsg, new=self.open_in_new_window, **keywords)

    @property
    def targets(self):
        """
        Returns a list of the ids of the running ds9 instances.
        """
        import ds9
        return ds9.ds9_targets()

    @property
    def current(self):
        """
        Return current ds9 instance.
        """
        targets = self.targets
        if targets is None:
            return None
        return self.get(targets[-1])

    def get(self, id):
        """
        Return ds9 instance matching a specified id.
        """
        import ds9
        return ds9.ds9(id)

    open_in_new_window = True

ds9 = Ds9()


def gaussian(shape, sigma=None, fwhm=None, origin=None, resolution=1,
             unit=None):
    """
    Returns an array whose values are the distances to a given origin.

    Parameters
    ----------
    shape : tuple of integer
        dimensions of the output array. For a 2d array, the first integer
        is for the Y-axis and the second one for the X-axis.
    fwhm : float or tuple of float
        The Full Width Half Maximum of the gaussian (fwhm_x, fwhm_y, ...).
    sigma : float or tuple of float
        The sigma parameter (sigma_x, sigma_y, ...).
    origin : array-like in the form of (x0, y0, ...), optional
        The coordinates of the point from which the distance is calculated,
        assuming a zero-based coordinate indexing. Default value is the array
        center.
    resolution : array-like in the form of (dx, dy, ...), optional
        Inter-pixel distance. Default is one.
    unit : string or dict
        The unit of the returned Map.

    """
    if sigma is None and fwhm is None:
        raise ValueError('The shape of the gaussian is not specified.')
    if sigma is None:
        sigma = fwhm / np.sqrt(8 * np.log(2))
    d = distance(shape, origin=origin, resolution=resolution / (np.sqrt(2) *
                 sigma))
    d.unit = ''
    d = np.exp(-d**2)
    d /= np.sum(d)
    if unit:
        d.unit = unit
    return d


def integrated_profile(input, origin=None, bin=1., nbins=None):
    """
    Returns axisymmetric integrated profile of a 2d image.
    x, y = integrated_profile(image, [origin, bin, nbins, histogram])

    Parameters
    ----------
    input: array
        2d input array.
    origin: (x0,y0)
        center of the profile. (Fits convention). Default is the image center.
    bin: number
        width of the profile bins (in unit of pixels).
    nbins: integer
        number of profile bins.

    Returns
    -------
    x : array
        The strict upper boundary within which each integration is performed.
    y : array
        The integrated profile.

    """
    x, y, n = profile(input, origin=origin, bin=bin, nbins=nbins,
                      histogram=True)
    x = np.arange(1, y.size + 1) * bin
    y[~np.isfinite(y)] = 0
    y *= n
    return x, np.add.accumulate(y)


def phasemask_fourquadrant(shape, phase=-1):
    array = Map.ones(shape, dtype=complex)
    array[0:shape[0]//2, shape[1]//2:] = phase
    array[shape[0]//2:, 0:shape[1]//2] = phase
    return array


def plot_tod(tod, mask=None, **kw):
    """Plot the signal timelines in a Tod and show masked samples.

    Plotting every detector timelines may be time consuming, so it is
    recommended to use this method on one or few detectors like this:
    >>> plot_tod(tod[idetector])

    """
    import matplotlib.pyplot as mp
    if mask is None:
        mask = getattr(tod, 'mask', None)

    ndetectors = product(tod.shape[0:-1])
    tod = tod.view().reshape((ndetectors, -1))
    if mask is not None:
        mask = mask.view().reshape((ndetectors, -1))
        if np.all(mask):
            print('There is no valid sample.')
            return

    for idetector in range(ndetectors):
        mp.plot(tod[idetector], **kw)
        if mask is not None:
            index = np.where(mask[idetector])
            mp.plot(index, tod[idetector, index], 'ro')

    unit = getattr(tod, 'unit', '')
    if unit:
        mp.ylabel('Signal [' + unit + ']')
    else:
        mp.ylabel('Signal')
    mp.xlabel('Time sample')


def profile(input, origin=None, bin=1., nbins=None, histogram=False):
    """
    Returns axisymmetric profile of a 2d image.
    x, y[, n] = profile(image, [origin, bin, nbins, histogram])

    Parameters
    ----------
    input: array
        2d input array
    origin: (x0,y0)
        Center of the profile. (Fits convention). Default is the image center.
    bin: number
        Width of the profile bins (in unit of pixels).
    nbins: integer
        Number of profile bins.
    histogram: boolean
        If set to True, return the histogram.

    """
    input = np.ascontiguousarray(input, float)
    if origin is None:
        origin = (np.array(input.shape[::-1], float) + 1) / 2
    else:
        origin = np.ascontiguousarray(origin, float)

    if nbins is None:
        nbins = int(max(input.shape[0]-origin[1], origin[1],
                        input.shape[1]-origin[0], origin[0]) / bin)

    x, y, n = flib.datautils.profile_axisymmetric_2d(input.T, origin, bin,
                                                     nbins)
    if histogram:
        return x, y, n
    else:
        return x, y


def psd2(array, sampling_frequency=1, fftw_flag='FFTW_MEASURE'):
    """
    Return two-dimensional PSD.

    Parameters
    ----------
    array : array-like
       Two-dimensional array.
    sampling_frequency : float
       The sampling frequency.
    fftw_flag : string
       The FFTW planner flag.

    Returns
    -------
    psd : ndarray
       The Power Spectrum Density, such as
           sum psd * bandwidth = mean(array**2)
       with bandwidth = sampling_frequency ** 2 / array.size is the PSD bin
       area.

    """
    array = np.asarray(array)
    s = np.abs(FFTOperator(array.shape, fftw_flag=fftw_flag)(array))**2
    s /= array.size**2
    bandwidth = sampling_frequency**2 / array.size
    s /= bandwidth
    return Map(scipy.fftpack.fftshift(s))


def profile_psd2(array, sampling_frequency=1, fftw_flag='FFTW_MEASURE'):
    """
    Return the axisymmetric profile of the PSD of a 2-dimensional image.

    Parameters
    ----------
    array : array-like
       Two-dimensional input array.
    sampling_frequency : float
       The sampling frequency.
    fftw_flag : string
       The FFTW planner flag.

    Returns
    -------
    x : ndarray
       The frequency bin centers, spaced by sampling_frequency / array.shape[1]
    y : ndarray
       The PSD profile. For white noise of standard deviation sigma, the
       following relationship stands:
           np.mean(y) * sampling_frequency**2 ~= sigma**2

    """
    array = np.asanyarray(array)
    if array.ndim != 2:
        raise ValueError('The input array is not two-dimensional.')
    if array.shape[0] != array.shape[1]:
        raise ValueError('The input array is not square.')
    psd = psd2(array, sampling_frequency=sampling_frequency,
               fftw_flag=fftw_flag)
    x, y = profile(psd, origin=array.shape // np.array(2) + 1)
    x *= sampling_frequency / array.shape[1]
    return x, y



from pyoperators import I
from pyoperators.utils.testing import assert_raises, assert_same
from pysimulators import Instrument, Imager, PackedTable


def test_instrument():

    def func(cls, keywords):
        instrument = cls(name, PackedTable(shape), **keywords)
        assert instrument.name == name
        assert instrument.detector.shape == shape
        assert len(instrument.detector) == len(instrument.detector.all)
        assert instrument

    name = 'instrument'
    shape = (3, 2)
    yield func, Instrument, {}
    yield func, Imager, {'object2image': I}
    yield func, Imager, {'image2object': I}


def test_error():
    assert_raises(ValueError, Imager, 'instrument', (3, 2))


def test_pack_unpack():
    layout = PackedTable(4, selection=[True, False, True, True])
    instrument = Instrument('instrument', layout)
    v = [1, 2, 3, 4]
    assert_same(instrument.pack(v), [1, 3, 4])
    assert_same(instrument.unpack([1, 3, 4]), [1, -1, 3, 4])



from __future__ import absolute_import, division, print_function

from pyoperators import MPI
from pysimulators import Acquisition, Instrument, PackedTable, Sampling, Scene

rank = MPI.COMM_WORLD.rank
size = MPI.COMM_WORLD.size


def test():
    scene = Scene(1024)
    instrument = Instrument('instrument', PackedTable((32, 32)))
    sampling = Sampling(1000)
    acq = Acquisition(instrument, sampling, scene,
                      nprocs_sampling=max(size // 2, 1))
    print(acq.comm.rank,
          acq.instrument.detector.comm.rank, '/',
          acq.instrument.detector.comm.size, acq.sampling.comm.rank, '/',
          acq.sampling.comm.size)



from __future__ import division

import numpy as np
import operator
from numpy.testing import assert_equal
from pyoperators import MaskOperator, PackOperator, Rotation2dOperator
from pyoperators.core import DeletedOperator
from pyoperators.utils.testing import assert_same, assert_is_type
from pysimulators.operators import ProjectionOperator
from pysimulators.sparse import (
    FSRMatrix, FSRRotation2dMatrix, FSRRotation3dMatrix)

clss = FSRMatrix, FSRRotation2dMatrix, FSRRotation3dMatrix
ftypes = np.float16, np.float32, np.float64, np.float128
itypes = np.int8, np.int16, np.int32, np.int64
index1 = [2, -1, 2, 1, 3, -1]
index2 = [-1, 3, 1, 1, 0, -1]
value1 = [0, 0, 0, 1.1, 0.3, 10]
value2 = [1, 2.2, 0.5, 1, 1, 10]
rotation1 = Rotation2dOperator(np.arange(6) * 30, degrees=True,
                               dtype=np.float128)
rotation2 = Rotation2dOperator(np.arange(6, 12) * 30, degrees=True,
                               dtype=np.float128)


def min_dtype(t1, t2):
    return min(t1, t2, key=lambda x: x().itemsize)


def _get_projection_fsr(itype, ftype, stype=None):
    if stype is None:
        stype = ftype
    dtype = [('index', itype), ('value', ftype)]
    data = np.recarray((6, 2), dtype=dtype)
    data[..., 0].index, data[..., 1].index = index1, index2
    data[..., 0].value, data[..., 1].value = value1, value2
    return ProjectionOperator(FSRMatrix((6, 5), data=data), dtype=stype)


def _get_projection_fsrrot2d(itype, ftype, stype=None):
    if stype is None:
        stype = ftype
    dtype = [('index', itype), ('r11', ftype), ('r21', ftype)]
    data = np.recarray((6, 2), dtype=dtype)
    data[..., 0].index, data[..., 1].index = index1, index2
    data[..., 0].r11 = value1 * rotation1.data[:, 0, 0]
    data[..., 1].r11 = value2 * rotation2.data[:, 0, 0]
    data[..., 0].r21 = value1 * rotation1.data[:, 1, 0]
    data[..., 1].r21 = value2 * rotation2.data[:, 1, 0]
    return ProjectionOperator(FSRRotation2dMatrix((6 * 2, 5 * 2), data=data),
                              dtype=stype)


def _get_projection_fsrrot3d(itype, ftype, stype=None):
    if stype is None:
        stype = ftype
    dtype = [('index', itype), ('r11', ftype), ('r22', ftype), ('r32', ftype)]
    data = np.recarray((6, 2), dtype=dtype)
    data[..., 0].index, data[..., 1].index = index1, index2
    data[..., 0].r11, data[..., 1].r11 = value1, value2
    data[..., 0].r22 = value1 * rotation1.data[:, 0, 0]
    data[..., 1].r22 = value2 * rotation2.data[:, 0, 0]
    data[..., 0].r32 = value1 * rotation1.data[:, 1, 0]
    data[..., 1].r32 = value2 * rotation2.data[:, 1, 0]
    return ProjectionOperator(FSRRotation3dMatrix((6 * 3, 5 * 3), data=data),
                              dtype=stype)

_get_projection = {
    FSRMatrix: _get_projection_fsr,
    FSRRotation2dMatrix: _get_projection_fsrrot2d,
    FSRRotation3dMatrix: _get_projection_fsrrot3d,
}


def test_kernel():
    expected = [False, False, True, False, True]

    def func(cls, itype, ftype):
        proj = _get_projection[cls](itype, ftype)
        actual = proj.canonical_basis_in_kernel()
        assert_equal(actual, expected)
        out = np.empty(5, bool)
        actual = proj.canonical_basis_in_kernel(out=out)
        assert_equal(actual, expected)
        kernel0 = [True, True, True, False, False]
        out = np.array(kernel0)
        proj.canonical_basis_in_kernel(out=out, operation=operator.iand)
        assert_equal(out, np.array(kernel0) & expected)
        kernel0 = [False, True, False, False, True]
        out = np.array(kernel0)
        proj.canonical_basis_in_kernel(out=out, operation=operator.ior)
        assert_equal(out, np.array(kernel0) | expected)
    for cls in clss:
        for itype in itypes:
            for ftype in ftypes:
                yield func, cls, itype, ftype


def test_pT1():
    def func(cls, itype, ftype, vtype):
        expected_ = np.asarray(expected, min_dtype(ftype, vtype))
        proj = _get_projection[cls](itype, ftype, vtype)
        pT1 = proj.pT1()
        assert_same(pT1, expected_)
        pT1 = proj.pT1(out=pT1)
        assert_same(pT1, expected_)
        pT1[...] = 2
        proj.pT1(out=pT1, operation=operator.iadd)
        assert_same(pT1, 2 + expected_)

    proj_ref = _get_projection[FSRMatrix](np.int32, np.float128)
    expected = proj_ref.T(np.ones(proj_ref.matrix.shape[0], np.float128))
    for cls in clss:
        for itype in itypes:
            for ftype in ftypes:
                for vtype in ftypes:
                    yield func, cls, itype, ftype, vtype


def test_pTx_pT1():
    input_fsr = np.array([1.1, 2, -1.3, 2, -3, 4.1], dtype=np.float128)
    input_fsrrot2d = input_fsr[:, None] * \
                     Rotation2dOperator(np.arange(6) * 30,
                                        dtype=np.float128)([1, 0])
    input_fsrrot3d = np.array([input_fsr, np.random.random_sample(6),
                               np.random.random_sample(6)]).T
    inputs = {FSRMatrix: input_fsr,
              FSRRotation2dMatrix: input_fsrrot2d,
              FSRRotation3dMatrix: input_fsrrot3d}

    def func(cls, itype, ftype, vtype):
        proj = _get_projection[cls](itype, ftype, vtype)
        input = np.asarray(inputs[cls], vtype)
        expectedx_ = np.asarray(expectedx, min_dtype(ftype, vtype))
        expected1_ = np.asarray(expected1, min_dtype(ftype, vtype))
        pTx, pT1 = proj.pTx_pT1(input)
        assert_same(pTx, expectedx_)
        assert_same(pT1, expected1_)
        proj.pTx_pT1(input, out=(pTx, pT1))
        assert_same(pTx, expectedx_)
        assert_same(pT1, expected1_)
        pTx[...] = 1
        pT1[...] = 2
        proj.pTx_pT1(input, out=(pTx, pT1), operation=operator.iadd)
        assert_same(pTx, 1 + expectedx_)
        assert_same(pT1, 2 + expected1_)

    proj_ref = _get_projection[FSRMatrix](np.int32, np.float128)
    expectedx = proj_ref.T(input_fsr)
    expected1 = proj_ref.T(np.ones(proj_ref.matrix.shape[0], np.float128))
    for cls in clss[0], clss[2]:
        for itype in itypes:
            for ftype in ftypes:
                for vtype in ftypes:
                    yield func, cls, itype, ftype, vtype


def test_restrict():
    restriction = np.array([True, False, False, True, True])
    kernel = [False, False, True, False, True]

    def func(cls, itype, ftype, inplace):
        proj_ref = _get_projection[cls](itype, ftype)
        proj_ = _get_projection[cls](itype, ftype)
        proj = proj_.restrict(restriction, inplace=inplace)
        if inplace:
            assert_is_type(proj_, DeletedOperator)
        if cls is not FSRMatrix:
            def pack(v):
                return v[restriction, :]
        else:
            pack = PackOperator(restriction)
        masking = MaskOperator(~restriction, broadcast='rightward')
        block_size = proj_ref.matrix.block_shape[1]
        shape = (5,) + ((block_size,) if block_size > 1 else ())
        x = np.arange(5*block_size).reshape(shape) + 1
        assert_equal(proj_ref(masking(x)), proj(pack(x)))
        pack = PackOperator(restriction)
        assert_equal(pack(kernel), proj.canonical_basis_in_kernel())
    for cls in clss:
        for itype in itypes:
            for ftype in ftypes:
                for inplace in False, True:
                    yield func, cls, itype, ftype, inplace



from __future__ import division

import itertools
import numpy as np
import pyoperators
import scipy.sparse
from numpy.testing import assert_equal, assert_raises
from pyoperators import (
    CompositionOperator, DiagonalOperator, Rotation2dOperator,
    Rotation3dOperator)
from pyoperators.utils.testing import assert_is_type, assert_same
from pysimulators.sparse import (
    FSCMatrix, FSRMatrix,
    FSCBlockMatrix, FSRBlockMatrix,
    FSCRotation2dMatrix, FSRRotation2dMatrix,
    FSCRotation3dMatrix, FSRRotation3dMatrix, SparseOperator)

ftypes = np.float16, np.float32, np.float64, np.float128
iutypes = (np.uint8, np.uint16, np.uint32, np.uint64)
itypes = (np.int8, np.int16, np.int32, np.int64)


def min_dtype(t1, t2):
    return min(t1, t2, key=lambda x: x().itemsize)


def test_fsc1():
    input = [1, 2, 1, 1, 1, 1]
    index = [3, 2, 2, 1, 2, -1]
    index_u = [3, 2, 2, 1, 2, 3]
    value = [1, 1, 0.5, 1, 2, 10]
    expected = [0, 1, 4.5, 1]
    expected_u = [0, 1, 4.5, 11]

    def func(itype, ftype, vtype, block_size):
        if np.dtype(itype).kind != 'u':
            ind = index
            exp = expected
        else:
            ind = index_u
            exp = expected_u
        input_ = np.array(input, vtype)
        if block_size == 2:
            input_ = np.array([input_, input_]).T.ravel()
            exp = np.array([exp, exp]).T.ravel()
        dtype = [('index', itype), ('value', ftype)]
        matrix = np.recarray((6, 1), dtype=dtype)
        matrix[..., 0].index = ind
        matrix[..., 0].value = value
        op = FSCMatrix((4, 6), data=matrix)
        out = op * np.array(input_, vtype)
        assert_same(out, exp)
        out[...] = 0
        op._matvec(np.array(input_, vtype), out=out)
        assert_same(out, exp)
    for itype in iutypes + itypes:
        for ftype in ftypes:
            for vtype in ftypes:
                for block_size in (1, 2):
                    yield func, itype, ftype, vtype, block_size


def test_fsc2():
    input_fsc = [1, 2, 1, 1, 1, 1]
    input_fsr = [1, 2, 1, 4]
    index1 = [3, 2, 2, 1, 2, -1]
    index2 = [-1, 3, 1, 1, 0, -1]
    index1_u = [3, 2, 2, 1, 2, 3]
    index2_u = [3, 3, 1, 1, 0, 3]
    value1 = [1, 1, 0.5, 1, 2, 10]
    value2 = [1, 2, 0.5, 1, 1, 10]
    expected = [1, 2.5, 4.5, 5]
    expected_u = [1, 2.5, 4.5, 26]

    def get_mat(itype, ftype):
        if np.dtype(itype).kind != 'u':
            ind = index1, index2
        else:
            ind = index1_u, index2_u
        dtype = [('index', itype), ('value', ftype)]
        matrix = np.recarray((6, 2), dtype=dtype)
        matrix[..., 0].index, matrix[..., 1].index = ind
        matrix[..., 0].value, matrix[..., 1].value = value1, value2
        return FSCMatrix((4, 6), data=matrix)

    def func1(itype, ftype, vtype, block_size):
        input_fsc_ = np.asarray(input_fsc, vtype)
        input_fsr_ = np.asarray(input_fsr, vtype)
        if np.dtype(itype).kind != 'u':
            exp = np.asarray(expected)
        else:
            exp = np.asarray(expected_u)
        if block_size == 2:
            input_fsc_ = np.array([input_fsc_, input_fsc_]).T.ravel()
            input_fsr_ = np.array([input_fsr_, input_fsr_]).T.ravel()
            exp = np.array([exp, exp]).T.ravel()
        mat = get_mat(itype, ftype)
        out = mat * input_fsc_
        assert_same(out, exp)
        out[...] = 0
        mat._matvec(input_fsc_, out=out)
        assert_same(out, exp)
        out = input_fsr_ * mat
        assert_same(out, FSRMatrix(mat.shape[::-1], data=mat.data) * 
                    input_fsr_)
        out = (3 * mat) * input_fsc_
        assert_same(out, 3 * exp)
        out = (mat * 3) * input_fsc_
        assert_same(out, 3 * exp)
    for itype in iutypes + itypes:
        for ftype in ftypes:
            for vtype in ftypes:
                for block_size in (1, 2):
                    yield func1, itype, ftype, vtype, block_size

    def func2(itype, ftype, shapein):
        mat = get_mat(itype, ftype)
        op = SparseOperator(mat)
        todense = op.todense(shapein=(6,) + shapein)
        assert_same(todense.T, op.T.todense(shapeout=(6,) + shapein))
        op2 = SparseOperator(mat, shapeout=(2, 2) + shapein,
                             shapein=(3, 2) + shapein)
        assert_same(op2.todense(), todense)
    for itype in iutypes + itypes:
        for ftype in ftypes:
            for shapein in (), (3,):
                yield func2, itype, ftype, shapein


def test_fsc3():
    mat = FSCMatrix((3, 4), 10)
    assert_equal(mat.data.shape, (4, 10))
    assert_equal(mat.data.index.dtype, int)
    assert_equal(mat.data.value.dtype, float)

    mat = FSCMatrix((3, 4), 10, dtype=np.float16, dtype_index=np.int8)
    assert_equal(mat.data.index.dtype, np.int8)
    assert_equal(mat.data.value.dtype, np.float16)


def test_fsc_error():
    data = np.zeros((3, 4), [('index', int), ('value', float)])
    data_wrong1 = np.empty((3, 4), [('index_', int), ('value', float)])
    data_wrong2 = np.empty((3, 4), [('value', float), ('index', int)])
    assert_raises(TypeError, FSCMatrix, 3)
    assert_raises(ValueError, FSCMatrix, (2, 3, 4))
    assert_raises(ValueError, FSCMatrix, (3, 8), data=data)
    assert_raises(TypeError, FSCMatrix, (8, 3), data=data_wrong1)
    assert_raises(TypeError, FSCMatrix, (8, 3), data=data_wrong2)
    assert_raises(ValueError, FSCMatrix, (8, 3), 5, data=data)
    FSCMatrix((8, 3), 4, data=data)
    assert_raises(ValueError, FSCMatrix, (8, 3))
    mat = FSCMatrix((8, 3), data=data)
    mat._matvec(np.ones(3))
    assert_raises(ValueError, mat._matvec, np.ones(7))
    assert_raises(TypeError, mat._matvec, np.ones(3), out=1)
    assert_raises(ValueError, mat._matvec, np.ones(3), out=np.zeros(7))


def test_fsr1():
    input = [1, 2, 3, 4]
    index = [3, 2, 2, 1, 2, -1]
    index_u = [3, 2, 2, 1, 2, 3]
    value = [1, 1, 0.5, 1, 2, 10]
    expected = [4, 3, 1.5, 2, 6, 0]
    expected_u = [4, 3, 1.5, 2, 6, 40]

    def get_mat(itype, ftype):
        if np.dtype(itype).kind != 'u':
            ind = index
        else:
            ind = index_u
        dtype = [('index', itype), ('value', ftype)]
        matrix = np.recarray((6, 1), dtype=dtype)
        matrix[..., 0].index = ind
        matrix[..., 0].value = value
        return FSRMatrix((6, 4), data=matrix)

    def func1(itype, ftype, vtype, block_size):
        if np.dtype(itype).kind != 'u':
            exp = expected
        else:
            exp = expected_u
        input_ = np.array(input, vtype)
        if block_size == 2:
            input_ = np.array([input_, input_]).T.ravel()
            exp = np.array([exp, exp]).T.ravel()
        op = get_mat(itype, ftype)
        out = op * input_
        assert_same(out, exp)
        out[...] = 0
        op._matvec(input_, out=out)
        assert_same(out, exp)
    for itype in iutypes + itypes:
        for ftype in ftypes:
            for vtype in ftypes:
                for block_size in (1, 2):
                    yield func1, itype, ftype, vtype, block_size

    def func2(itype, ftype):
        mat = get_mat(itype, ftype)
        op = SparseOperator(mat, shapein=4)
        todense = op.todense()
        pTp = op.T * op
        if (itype, ftype) not in ((np.int32, np.float32),
                                  (np.int32, np.float64),
                                  (np.int64, np.float32),
                                  (np.int64, np.float64)):
            assert_is_type(pTp, CompositionOperator)
            return
        assert_is_type(pTp, DiagonalOperator)
        assert_same(pTp.todense(), np.dot(todense.T, todense))
    for itype in iutypes + itypes:
        for ftype in ftypes:
            yield func2, itype, ftype


def test_fsr2():
    input_fsr = [1, 2, 3, 4]
    input_fsc = [1, 2, 3, 4, 5, 6]
    index1 = [3, 2, 2, 1, 2, -1]
    index2 = [-1, 3, 1, 1, 0, -1]
    index1_u = [3, 2, 2, 1, 2, 3]
    index2_u = [3, 3, 1, 1, 0, 3]
    value1 = [1, 1, 0.5, 1, 2, 10]
    value2 = [1, 2, 0.5, 1, 1, 10]
    expected = [4, 11, 2.5, 4, 7, 0]
    expected_u = [8, 11, 2.5, 4, 7, 80]

    def get_mat(itype, ftype):
        if np.dtype(itype).kind != 'u':
            ind = index1, index2
        else:
            ind = index1_u, index2_u
        dtype = [('index', itype), ('value', ftype)]
        matrix = np.recarray((6, 2), dtype=dtype)
        matrix[..., 0].index, matrix[..., 1].index = ind
        matrix[..., 0].value, matrix[..., 1].value = value1, value2
        return FSRMatrix((6, 4), data=matrix)

    def func1(itype, ftype, vtype, block_size):
        input_fsc_ = np.asarray(input_fsc, vtype)
        input_fsr_ = np.asarray(input_fsr, vtype)
        if np.dtype(itype).kind != 'u':
            exp = np.asarray(expected)
        else:
            exp = np.asarray(expected_u)
        if block_size == 2:
            input_fsc_ = np.array([input_fsc_, input_fsc_]).T.ravel()
            input_fsr_ = np.array([input_fsr_, input_fsr_]).T.ravel()
            exp = np.array([exp, exp]).T.ravel()
        mat = get_mat(itype, ftype)
        out = mat * input_fsr_
        assert_same(out, exp)
        out[...] = 0
        mat._matvec(input_fsr_, out=out)
        assert_same(out, exp)
        out = input_fsc_ * mat
        assert_same(out, FSCMatrix(mat.shape[::-1], data=mat.data) * 
                    input_fsc_)
        out = (3 * mat) * input_fsr_
        assert_same(out, 3 * exp)
        out = (mat * 3) * input_fsr_
        assert_same(out, 3 * exp)
    for itype in iutypes + itypes:
        for ftype in ftypes:
            for vtype in ftypes:
                for block_size in (1, 2):
                    yield func1, itype, ftype, vtype, block_size

    def func2(itype, ftype, shapein):
        mat = get_mat(itype, ftype)
        op = SparseOperator(mat)
        todense = op.todense(shapein=(4,) + shapein)
        assert_same(todense.T, op.T.todense(shapeout=(4,) + shapein))
        if shapein is None:
            return
        op2 = SparseOperator(mat, shapein=(2, 2) + shapein,
                             shapeout=(3, 2) + shapein)
        assert_same(op2.todense(), todense)
    for itype in iutypes + itypes:
        for ftype in ftypes:
            for shapein in (), (3,):
                yield func2, itype, ftype, shapein


def test_fsr3():
    mat = FSRMatrix((4, 3), 10)
    assert_equal(mat.data.shape, (4, 10))
    assert_equal(mat.data.index.dtype, int)
    assert_equal(mat.data.value.dtype, float)

    mat = FSRMatrix((4, 3), 10, dtype=np.float16, dtype_index=np.int8)
    assert_equal(mat.data.index.dtype, np.int8)
    assert_equal(mat.data.value.dtype, np.float16)



def test_fsr_error():
    data = np.zeros((3, 4), [('index', int), ('value', float)])
    data_wrong1 = np.empty((3, 4), [('index_', int), ('value', float)])
    data_wrong2 = np.empty((3, 4), [('value', float), ('index', int)])
    assert_raises(TypeError, FSRMatrix, 3)
    assert_raises(ValueError, FSRMatrix, (2, 3, 4))
    assert_raises(ValueError, FSRMatrix, (8, 3), data=data)
    assert_raises(TypeError, FSRMatrix, (3, 8), data=data_wrong1)
    assert_raises(TypeError, FSRMatrix, (3, 8), data=data_wrong2)
    assert_raises(ValueError, FSRMatrix, (3, 8), 5, data=data)
    FSRMatrix((3, 8), 4, data=data)
    assert_raises(ValueError, FSRMatrix, (3, 8))
    mat = FSRMatrix((3, 8), data=data)
    mat._matvec(np.ones(8))
    assert_raises(ValueError, mat._matvec, np.ones(7))
    assert_raises(TypeError, mat._matvec, np.ones(8), out=1)
    assert_raises(ValueError, mat._matvec, np.ones(8), out=np.zeros(4))


def test_block():
    outer_block_shape = 6, 4
    index1 = [3, 2, 2, 1, 2, -1]
    index2 = [-1, 3, 1, 1, 0, -1]
    np.random.seed(0)

    def get_mat(itype, ftype, inner_block_shape, indices):
        shape = [i * o for i, o in zip(inner_block_shape, outer_block_shape)]
        mat = FSRBlockMatrix(shape, inner_block_shape, len(indices),
                             dtype=ftype, dtype_index=itype)
        for k, index in enumerate(indices):
            mat.data[:, k].index = index
        for i in range(outer_block_shape[0]):
            for k in range(len(indices)):
                mat.data[i, k]['value'][...] = np.random.random_integers(
                    -10, 10, inner_block_shape)
        dense = np.zeros(shape, dtype=ftype)
        for i, d in enumerate(mat.data):
            a1, z1 = i * inner_block_shape[0], (i+1) * inner_block_shape[0]
            for d_ in d:
                j = d_['index']
                if j < 0:
                    continue
                a2, z2 = j * inner_block_shape[1], (j+1) * inner_block_shape[1]
                dense[a1:z1, a2:z2] += d_['value']
        return mat, dense

    def func1(itype, ftype, inner_block_shape):
        if inner_block_shape == (1, 1):
            return
        mat_fsr, dense = get_mat(itype, ftype, inner_block_shape,
                                 [index1, index2])
        mat_fsc = mat_fsr._transpose()
        input_fsc = np.random.random_sample(mat_fsc.shape[1])
        input_fsr = np.random.random_sample(mat_fsr.shape[1])

        assert_same(mat_fsr.block_shape, inner_block_shape)
        assert_same(mat_fsc.block_shape, inner_block_shape[::-1])
        op = SparseOperator(mat_fsc)
        assert_equal(op.matrix.dtype, ftype)
        assert_equal(op.dtype, ftype)
        assert_same(op.todense(), dense.T)
        assert_same(op.T.todense(), dense)
        ref = mat_fsc._matvec(input_fsc.astype(np.float128))
        for ftype2 in ftypes[:-1]:
            out = np.zeros_like(ref, ftype2)
            mat_fsc._matvec(input_fsc.astype(ftype2), out)
            assert_same(out, ref.astype(min_dtype(ftype, ftype2)), atol=10)
        ref = (2 * (mat_fsc * input_fsc)).astype(ftype)
        assert_same((mat_fsc * 2) * input_fsc, ref, atol=10)
        assert_same((2 * mat_fsc) * input_fsc, ref, atol=10)
        assert_same(input_fsr * mat_fsc, mat_fsr * input_fsr, atol=10)

        op = SparseOperator(mat_fsr)
        assert_equal(op.matrix.dtype, ftype)
        assert_equal(op.dtype, ftype)
        assert_same(op.todense(), dense)
        assert_same(op.T.todense(), dense.T)
        ref = mat_fsr._matvec(input_fsr.astype(np.float128))
        for ftype2 in ftypes[:-1]:
            out = np.zeros_like(ref, ftype2)
            mat_fsr._matvec(input_fsr.astype(ftype2), out)
            assert_same(out, ref.astype(min_dtype(ftype, ftype2)), atol=10)
        ref = (2 * (mat_fsr * input_fsr)).astype(ftype)
        assert_same((mat_fsr * 2) * input_fsr, ref, atol=10)
        assert_same((2 * mat_fsr) * input_fsr, ref, atol=10)
        assert_same(input_fsc * mat_fsr, mat_fsc * input_fsc, atol=10)

    for itype in itypes:
        for ftype in ftypes:
            for block_shape in itertools.product([1, 2, 3], repeat=2):
                yield func1, itype, ftype, block_shape


def test_rot2d():
    input_fsc = np.arange(6*2.)
    input_fsr = np.arange(4*2.)
    index1 = [3, 2, 2, 1, 2, -1]
    index2 = [-1, 3, 1, 1, 0, -1]
    value1 = [1, 1, 0.5, 1, 2, 10]
    value2 = [1, 2, 0.5, 1, 1, 10]
    angle1 = [1, 10, 20, 30, 40, -10]
    angle2 = [0, 30, -20, 1, 1, 10]

    def fill(array, dense, index, value, angle, n):
        for i, (j, v, a) in enumerate(zip(index, value, angle)):
            array[i, n].index = j
            if j == -1:
                array[i, n].r11 = 0
                array[i, n].r21 = 0
                continue
            r = v * Rotation2dOperator(a, degrees=True).todense(shapein=2)
            array[i, n].r11 = r[0, 0]
            array[i, n].r21 = r[1, 0]
            dense[2*i:2*i+2, 2*j:2*j+2] += r

    def func1(itype, ftype):
        array = np.recarray((6, 2), dtype=[('index', itype), ('r11', ftype),
                                           ('r21', ftype)])
        dense = np.zeros((6*2, 4*2), dtype=ftype)
        fill(array, dense, index1, value1, angle1, 0)
        fill(array, dense, index2, value2, angle2, 1)

        mat_fsc = FSCRotation2dMatrix(dense.shape[::-1], data=array)
        mat_fsr = FSRRotation2dMatrix(dense.shape, data=array)

        op = SparseOperator(mat_fsc)
        assert_equal(op.matrix.dtype, ftype)
        assert_equal(op.dtype, ftype)
        assert_same(op.todense(), dense.T)
        assert_same(op.T.todense(), dense)
        ref = mat_fsc._matvec(input_fsc.astype(np.float128))
        for ftype2 in ftypes[:-1]:
            out = np.zeros_like(ref, ftype2)
            mat_fsc._matvec(input_fsc.astype(ftype2), out)
            assert_same(out, ref.astype(min_dtype(ftype, ftype2)))
        ref = (2 * (mat_fsc * input_fsc)).astype(ftype)
        assert_same((mat_fsc * 2) * input_fsc, ref)
        assert_same((2 * mat_fsc) * input_fsc, ref)
        assert_same(input_fsr * mat_fsc, mat_fsr * input_fsr)

        op = SparseOperator(mat_fsr)
        assert_equal(op.matrix.dtype, ftype)
        assert_equal(op.dtype, ftype)
        assert_same(op.todense(), dense)
        assert_same(op.T.todense(), dense.T)
        ref = mat_fsr._matvec(input_fsr.astype(np.float128))
        for ftype2 in ftypes[:-1]:
            out = np.zeros_like(ref, ftype2)
            mat_fsr._matvec(input_fsr.astype(ftype2), out)
            assert_same(out, ref.astype(min_dtype(ftype, ftype2)))
        ref = (2 * (mat_fsr * input_fsr)).astype(ftype)
        assert_same((mat_fsr * 2) * input_fsr, ref)
        assert_same((2 * mat_fsr) * input_fsr, ref)
        assert_same(input_fsc * mat_fsr, mat_fsc * input_fsc)

    for itype in itypes:
        for ftype in ftypes:
            yield func1, itype, ftype

    def func2(itype, ftype):
        array = np.recarray((6, 1), dtype=[('index', itype), ('r11', ftype),
                                           ('r21', ftype)])
        dense = np.zeros((6*2, 4*2), dtype=ftype)
        fill(array, dense, index1, value1, angle1, 0)
        op = SparseOperator(FSRRotation2dMatrix(dense.shape, data=array))
        pTp = op.T * op
        if (itype, ftype) not in ((np.int32, np.float32),
                                  (np.int32, np.float64),
                                  (np.int64, np.float32),
                                  (np.int64, np.float64)):
            assert_is_type(pTp, CompositionOperator)
            return
        assert_is_type(pTp, DiagonalOperator)
        assert_same(pTp.todense(), np.dot(dense.T, dense))

    for itype in itypes:
        for ftype in ftypes:
            yield func2, itype, ftype


def test_fsc_rot2d_error():
    data = np.zeros((3, 4), dtype=[('index', int), ('r11', float),
                                   ('r21', float)])
    data_wrong1 = np.empty((3, 4), [('index', int), ('value', float)])
    data_wrong2 = np.empty((3, 4), [('r11', float), ('r21', float),
                                    ('index', int)])
    data_wrong3 = np.empty((3, 4), dtype=[('index', int), ('r11', float),
                                          ('r21', np.float32)])
    assert_raises(TypeError, FSCRotation2dMatrix, 2)
    assert_raises(ValueError, FSCRotation2dMatrix, (2, 3, 4))
    assert_raises(ValueError, FSCRotation2dMatrix, (6, 16), data=data)
    assert_raises(ValueError, FSCRotation2dMatrix, (7, 16), data=data)
    assert_raises(ValueError, FSCRotation2dMatrix, (17, 6), data=data)
    assert_raises(TypeError, FSCRotation2dMatrix, (16, 6), data=data_wrong1)
    assert_raises(TypeError, FSCRotation2dMatrix, (16, 6), data=data_wrong2)
    assert_raises(TypeError, FSCRotation2dMatrix, (16, 6), data=data_wrong3)
    assert_raises(ValueError, FSCRotation2dMatrix, (16, 6), 5, data=data)
    assert_raises(ValueError, FSCRotation2dMatrix, (16, 6))
    mat = FSCRotation2dMatrix((16, 6), data=data)
    assert_raises(ValueError, mat._matvec, np.ones(4))
    mat._matvec(np.ones((3, 2)))
    assert_raises(ValueError, mat._matvec, np.ones(17))
    assert_raises(ValueError, mat._matvec, np.ones((8, 2)))
    assert_raises(TypeError, mat._matvec, np.ones((3, 2)), out=1)
    assert_raises(ValueError, mat._matvec, np.ones((3, 2)), out=np.zeros(7))


def test_fsr_rot2d_error():
    data = np.zeros((3, 4), dtype=[('index', int), ('r11', float),
                                   ('r21', float)])
    data_wrong1 = np.empty((3, 4), [('index', int), ('value', float)])
    data_wrong2 = np.empty((3, 4), [('r11', float), ('r21', float),
                                    ('index', int)])
    data_wrong3 = np.empty((3, 4), dtype=[('index', int), ('r11', float),
                                          ('r21', np.float32)])
    assert_raises(TypeError, FSRRotation2dMatrix, 3)
    assert_raises(ValueError, FSRRotation2dMatrix, (2, 3, 4))
    assert_raises(ValueError, FSRRotation2dMatrix, (16, 6), data=data)
    assert_raises(ValueError, FSRRotation2dMatrix, (7, 16), data=data)
    assert_raises(ValueError, FSRRotation2dMatrix, (6, 17), data=data)
    assert_raises(TypeError, FSRRotation2dMatrix, (6, 16), data=data_wrong1)
    assert_raises(TypeError, FSRRotation2dMatrix, (6, 16), data=data_wrong2)
    assert_raises(TypeError, FSRRotation2dMatrix, (6, 16), data=data_wrong3)
    assert_raises(ValueError, FSRRotation2dMatrix, (6, 16), 5, data=data)
    assert_raises(ValueError, FSRRotation2dMatrix, (6, 16))
    mat = FSRRotation2dMatrix((6, 16), data=data)
    assert_raises(ValueError, mat._matvec, np.ones(4))
    mat._matvec(np.ones((8, 2)))
    assert_raises(ValueError, mat._matvec, np.ones(17))
    assert_raises(ValueError, mat._matvec, np.ones((3, 2)))
    assert_raises(TypeError, mat._matvec, np.ones((8, 2)), out=1)
    assert_raises(ValueError, mat._matvec, np.ones((8, 2)), out=np.zeros(7))


def test_rot3d():
    input_fsc = np.arange(6*3.)
    input_fsr = np.arange(4*3.)
    index1 = [3, 2, 2, 1, 2, -1]
    index2 = [-1, 3, 1, 1, 0, -1]
    value1 = [1, 1, 0.5, 1, 2, 10]
    value2 = [1, 2, 0.5, 1, 1, 10]
    angle1 = [1, 10, 20, 30, 40, -10]
    angle2 = [0, 30, -20, 1, 1, 10]

    def fill(array, dense, index, value, angle, n):
        for i, (j, v, a) in enumerate(zip(index, value, angle)):
            array[i, n].index = j
            if j == -1:
                array[i, n].r11 = 0
                array[i, n].r22 = 0
                array[i, n].r32 = 0
                continue
            r = v * Rotation3dOperator('X', a, degrees=True).todense(shapein=3)
            array[i, n].r11 = r[0, 0]
            array[i, n].r22 = r[1, 1]
            array[i, n].r32 = r[2, 1]
            dense[3*i:3*i+3, 3*j:3*j+3] += r

    def func1(itype, ftype):
        array = np.recarray((6, 2), dtype=[('index', itype), ('r11', ftype),
                                           ('r22', ftype), ('r32', ftype)])
        dense = np.zeros((6*3, 4*3), dtype=ftype)
        fill(array, dense, index1, value1, angle1, 0)
        fill(array, dense, index2, value2, angle2, 1)

        mat_fsc = FSCRotation3dMatrix(dense.shape[::-1], data=array)
        mat_fsr = FSRRotation3dMatrix(dense.shape, data=array)

        op = SparseOperator(mat_fsc)
        assert_equal(op.matrix.dtype, ftype)
        assert_equal(op.dtype, ftype)
        assert_same(op.todense(), dense.T)
        assert_same(op.T.todense(), dense)
        ref = mat_fsc._matvec(input_fsc.astype(np.float128))
        for ftype2 in ftypes[:-1]:
            out = np.zeros_like(ref, ftype2)
            mat_fsc._matvec(input_fsc.astype(ftype2), out)
            assert_same(out, ref.astype(min_dtype(ftype, ftype2)))
        ref = (3 * (mat_fsc * input_fsc)).astype(ftype)
        assert_same((mat_fsc * 3) * input_fsc, ref)
        assert_same((3 * mat_fsc) * input_fsc, ref)
        assert_same(input_fsr * mat_fsc, mat_fsr * input_fsr)

        op = SparseOperator(mat_fsr)
        assert_equal(op.matrix.dtype, ftype)
        assert_equal(op.dtype, ftype)
        assert_same(op.todense(), dense)
        assert_same(op.T.todense(), dense.T)
        ref = mat_fsr._matvec(input_fsr.astype(np.float128))
        for ftype2 in ftypes[:-1]:
            out = np.zeros_like(ref, ftype2)
            mat_fsr._matvec(input_fsr.astype(ftype2), out)
            assert_same(out, ref.astype(min_dtype(ftype, ftype2)))
        ref = (3 * (mat_fsr * input_fsr)).astype(ftype)
        assert_same((mat_fsr * 3) * input_fsr, ref)
        assert_same((3 * mat_fsr) * input_fsr, ref)
        assert_same(input_fsc * mat_fsr, mat_fsc * input_fsc)

    for itype in itypes:
        for ftype in ftypes:
            yield func1, itype, ftype

    def func2(itype, ftype):
        array = np.recarray((6, 1), dtype=[('index', itype), ('r11', ftype),
                                           ('r22', ftype), ('r32', ftype)])
        dense = np.zeros((6*3, 4*3), dtype=ftype)
        fill(array, dense, index1, value1, angle1, 0)
        op = SparseOperator(FSRRotation3dMatrix(dense.shape, data=array))
        pTp = op.T * op
        if (itype, ftype) not in ((np.int32, np.float32),
                                  (np.int32, np.float64),
                                  (np.int64, np.float32),
                                  (np.int64, np.float64)):
            assert_is_type(pTp, CompositionOperator)
            return
        assert_is_type(pTp, DiagonalOperator)
        assert_same(pTp.todense(), np.dot(dense.T, dense), atol=1)

    for itype in itypes:
        for ftype in ftypes:
            yield func2, itype, ftype


def test_fsc_rot3d_error():
    data = np.zeros((3, 4), dtype=[('index', int), ('r11', float),
                                   ('r22', float), ('r32', float)])
    data_wrong1 = np.empty((3, 4), [('index', int), ('value', float)])
    data_wrong2 = np.empty((3, 4), [('r11', float), ('r22', float),
                                    ('r32', float), ('index', int)])
    data_wrong3 = np.empty((3, 4), dtype=[('index', int), ('r11', float),
                                          ('r22', np.float32), ('r32', float)])
    assert_raises(TypeError, FSCRotation3dMatrix, 3)
    assert_raises(ValueError, FSCRotation3dMatrix, (2, 3, 4))
    assert_raises(ValueError, FSCRotation3dMatrix, (9, 24), data=data)
    assert_raises(ValueError, FSCRotation3dMatrix, (10, 24), data=data)
    assert_raises(ValueError, FSCRotation3dMatrix, (25, 9), data=data)
    assert_raises(TypeError, FSCRotation3dMatrix, (24, 9), data=data_wrong1)
    assert_raises(TypeError, FSCRotation3dMatrix, (24, 9), data=data_wrong2)
    assert_raises(TypeError, FSCRotation3dMatrix, (24, 9), data=data_wrong3)
    assert_raises(ValueError, FSCRotation3dMatrix, (24, 9), 5, data=data)
    assert_raises(ValueError, FSCRotation3dMatrix, (24, 9))
    mat = FSCRotation3dMatrix((24, 9), data=data)
    assert_raises(ValueError, mat._matvec, np.ones(4))
    mat._matvec(np.ones((3, 3)))
    assert_raises(ValueError, mat._matvec, np.ones(25))
    assert_raises(ValueError, mat._matvec, np.ones((8, 3)))
    assert_raises(TypeError, mat._matvec, np.ones((3, 3)), out=1)
    assert_raises(ValueError, mat._matvec, np.ones((3, 3)), out=np.zeros(7))


def test_fsr_rot3d_error():
    data = np.zeros((3, 4), dtype=[('index', int), ('r11', float),
                                   ('r22', float), ('r32', float)])
    data_wrong1 = np.empty((3, 4), [('index', int), ('value', float)])
    data_wrong2 = np.empty((3, 4), [('r11', float), ('r22', float),
                                    ('r32', float), ('index', int)])
    data_wrong3 = np.empty((3, 4), dtype=[('index', int), ('r11', float),
                                          ('r22', np.float32), ('r32', float)])
    assert_raises(TypeError, FSRRotation3dMatrix, 3)
    assert_raises(ValueError, FSRRotation3dMatrix, (2, 3, 4))
    assert_raises(ValueError, FSRRotation3dMatrix, (24, 9), data=data)
    assert_raises(ValueError, FSRRotation3dMatrix, (10, 24), data=data)
    assert_raises(ValueError, FSRRotation3dMatrix, (9, 25), data=data)
    assert_raises(TypeError, FSRRotation3dMatrix, (9, 24), data=data_wrong1)
    assert_raises(TypeError, FSRRotation3dMatrix, (9, 24), data=data_wrong2)
    assert_raises(TypeError, FSRRotation3dMatrix, (9, 24), data=data_wrong3)
    assert_raises(ValueError, FSRRotation3dMatrix, (9, 24), 5, data=data)
    assert_raises(ValueError, FSRRotation3dMatrix, (9, 24))
    mat = FSRRotation3dMatrix((9, 24), data=data)
    assert_raises(ValueError, mat._matvec, np.ones(4))
    mat._matvec(np.ones((8, 3)))
    assert_raises(ValueError, mat._matvec, np.ones(25))
    assert_raises(ValueError, mat._matvec, np.ones((3, 3)))
    assert_raises(TypeError, mat._matvec, np.ones((8, 3)), out=1)
    assert_raises(ValueError, mat._matvec, np.ones((8, 3)), out=np.zeros(7))


def test_sparse_operator():
    assert_raises(TypeError, SparseOperator, np.zeros((10, 3)))
    sp = SparseOperator(scipy.sparse.csc_matrix((4, 6)))
    assert_is_type(sp, pyoperators.linear.SparseOperator)

    data = FSRMatrix((10, 3), ncolmax=1)
    assert_raises(ValueError, SparseOperator, data, shapein=4)
    assert_raises(ValueError, SparseOperator, data, shapeout=11)

    data = FSRRotation2dMatrix((12, 6), ncolmax=1)
    assert_raises(ValueError, SparseOperator, data, shapein=(5, 2))
    assert_raises(ValueError, SparseOperator, data, shapeout=(7,))

    data = FSRRotation3dMatrix((12, 6), ncolmax=1)
    assert_raises(ValueError, SparseOperator, data, shapein=(5, 3))
    assert_raises(ValueError, SparseOperator, data, shapeout=(7,))
